# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for `components` command group in KFP CLI."""
import contextlib
import functools
import os
import pathlib
import subprocess
import textwrap
from typing import List, Optional, Union
import unittest
from unittest import mock

from click import testing
import docker

docker = mock.Mock()

from kfp.cli import component


def _make_component(
    func_name: str,
    base_image: Optional[str] = None,
    target_image: Optional[str] = None,
    packages_to_install: Optional[List[str]] = None,
    output_component_file: Optional[str] = None,
    pip_index_urls: Optional[List[str]] = None,
) -> str:
    return textwrap.dedent('''
    from kfp.dsl import *

    @component(
        base_image={base_image},
        target_image={target_image},
        packages_to_install={packages_to_install},
        output_component_file={output_component_file},
        pip_index_urls={pip_index_urls})
    def {func_name}():
        pass
    ''').format(
        base_image=repr(base_image),
        target_image=repr(target_image),
        packages_to_install=repr(packages_to_install),
        output_component_file=repr(output_component_file),
        pip_index_urls=repr(pip_index_urls),
        func_name=func_name)


def _write_file(filename: str, file_contents: str):
    filepath = pathlib.Path(filename)
    filepath.parent.mkdir(exist_ok=True, parents=True)
    filepath.write_text(file_contents)


def _write_components(filename: str, component_template: Union[List[str], str]):
    if isinstance(component_template, list):
        file_contents = '\n\n'.join(component_template)
    else:
        file_contents = component_template
    _write_file(filename=filename, file_contents=file_contents)


class Test(unittest.TestCase):

    def setUp(self) -> None:
        self.runner = testing.CliRunner()
        self.cli = component.component
        component._DOCKER_IS_PRESENT = True

        patcher = mock.patch('docker.from_env')
        self._docker_client = patcher.start().return_value
        self._docker_client.images.build.return_value = [{
            'stream': 'Build logs'
        }]
        self._docker_client.images.push.return_value = [{'status': 'Pushed'}]
        self.addCleanup(patcher.stop)

        with contextlib.ExitStack() as stack:
            stack.enter_context(self.runner.isolated_filesystem())
            self._working_dir = pathlib.Path.cwd()
            self.addCleanup(stack.pop_all().close)

        return super().setUp()

    def assert_file_exists(self, path: str):
        path_under_test_dir = self._working_dir / path
        self.assertTrue(path_under_test_dir, f'File {path} does not exist!')

    def assert_file_exists_and_contains(self, path: str, expected_content: str):
        self.assert_file_exists(path)
        path_under_test_dir = self._working_dir / path
        got_content = path_under_test_dir.read_text()
        self.assertEqual(got_content, expected_content)

    def test_kfp_config_for_single_file(self):
        preprocess_component = _make_component(
            func_name='preprocess', target_image='custom-image')
        train_component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py',
                          [preprocess_component, train_component])

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)

        self.assert_file_exists_and_contains(
            'kfp_config.ini',
            textwrap.dedent('''\
            [Components]
            preprocess = components.py
            train = components.py

            '''))

    def test_kfp_config_for_single_file_under_nested_directory(self):
        preprocess_component = _make_component(
            func_name='preprocess', target_image='custom-image')
        train_component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('dir1/dir2/dir3/components.py',
                          [preprocess_component, train_component])

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)

        self.assert_file_exists_and_contains(
            'kfp_config.ini',
            textwrap.dedent('''\
            [Components]
            preprocess = dir1/dir2/dir3/components.py
            train = dir1/dir2/dir3/components.py

            '''))

    def test_kfp_config_for_multiple_files(self):
        component = _make_component(
            func_name='preprocess', target_image='custom-image')
        _write_components('preprocess_component.py', component)

        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('train_component.py', component)

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)

        self.assert_file_exists_and_contains(
            'kfp_config.ini',
            textwrap.dedent('''\
            [Components]
            preprocess = preprocess_component.py
            train = train_component.py

            '''))

    def test_kfp_config_for_multiple_files_under_nested_directories(self):
        component = _make_component(
            func_name='preprocess', target_image='custom-image')
        _write_components('preprocess/preprocess_component.py', component)

        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('train/train_component.py', component)

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)

        self.assert_file_exists_and_contains(
            'kfp_config.ini',
            textwrap.dedent('''\
            [Components]
            preprocess = preprocess/preprocess_component.py
            train = train/train_component.py

            '''))

    def test_target_image_must_be_the_same_in_all_components(self):
        component_one = _make_component(func_name='one', target_image='image-1')
        component_two = _make_component(func_name='two', target_image='image-1')
        _write_components('one_two/one_two.py', [component_one, component_two])

        component_three = _make_component(
            func_name='three', target_image='image-2')
        component_four = _make_component(
            func_name='four', target_image='image-3')
        _write_components('three_four/three_four.py',
                          [component_three, component_four])

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 1)

    def test_target_image_must_be_the_same_in_all_components_with_base_image(
            self):
        component_one = _make_component(
            func_name='one', base_image='image-1', target_image='target-image')
        component_two = _make_component(
            func_name='two', base_image='image-1', target_image='target-image')
        _write_components('one_two/one_two.py', [component_one, component_two])

        component_three = _make_component(
            func_name='three',
            base_image='image-2',
            target_image='target-image')
        component_four = _make_component(
            func_name='four', base_image='image-3', target_image='target-image')
        _write_components('three_four/three_four.py',
                          [component_three, component_four])

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 1)

    def test_component_filepattern_can_be_used_to_restrict_discovery(self):
        component = _make_component(
            func_name='preprocess', target_image='custom-image')
        _write_components('preprocess/preprocess_component.py', component)

        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('train/train_component.py', component)

        result = self.runner.invoke(
            self.cli,
            [
                'build',
                str(self._working_dir), '--component-filepattern=train/*'
            ],
        )
        self.assertEqual(result.exit_code, 0)

        self.assert_file_exists_and_contains(
            'kfp_config.ini',
            textwrap.dedent('''\
            [Components]
            train = train/train_component.py

            '''))

    def test_nested_module_imports(self):
        module_two = 'two = 2'
        _write_file('module_two.py', module_two)

        module_one = 'from module_two import two\none = 1'
        _write_file('module_one.py', module_one)

        component = _make_component(
            func_name='comp', target_image='custom-image')
        component = 'from module_one import one\n' + component
        _write_components('component.py', component)

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)

    def test_emptry_requirements_txt_file_is_generated(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        result = self.runner.invoke(self.cli, ['build', str(self._working_dir)])
        self.assertEqual(result.exit_code, 0)
        self.assert_file_exists_and_contains('runtime-requirements.txt',
                                             '# Generated by KFP.\n')

    def test_existing_requirements_txt_file_is_updated(self):
        component = _make_component(
            func_name='train',
            target_image='custom-image',
            packages_to_install=[
                'tensorflow==2.10.0', 'google-cloud-pipeline-components'
            ])
        _write_components('components.py', component)

        _write_file('runtime-requirements.txt', 'Some pre-existing content')

        result = self.runner.invoke(self.cli, ['build', str(self._working_dir)])
        self.assertEqual(result.exit_code, 0)
        self.assert_file_exists_and_contains(
            'runtime-requirements.txt',
            '# Generated by KFP.\ngoogle-cloud-pipeline-components\ntensorflow==2.10.0'
        )

    def test_user_requirements_txt_file_is_unchanged(self):
        component = _make_component(
            func_name='train',
            target_image='custom-image',
            packages_to_install=['tensorflow==2.10.0'])
        _write_components('components.py', component)

        _write_file('requirements.txt', 'Some pre-existing content')

        result = self.runner.invoke(self.cli, ['build', str(self._working_dir)])
        self.assertEqual(result.exit_code, 0)
        self.assert_file_exists_and_contains('requirements.txt',
                                             'Some pre-existing content')

    def test_docker_ignore_file_is_generated(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        result = self.runner.invoke(self.cli, ['build', str(self._working_dir)])
        self.assertEqual(result.exit_code, 0)
        self.assert_file_exists_and_contains(
            '.dockerignore',
            textwrap.dedent('''\
            # Generated by KFP.

            component_metadata/
            '''))

    def test_existing_docker_ignore_file_is_unchanged(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        _write_file('.dockerignore', 'Some pre-existing content')

        result = self.runner.invoke(self.cli, ['build', str(self._working_dir)])
        self.assertEqual(result.exit_code, 0)
        self.assert_file_exists_and_contains('.dockerignore',
                                             'Some pre-existing content')

    def test_docker_engine_is_supported(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)
        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir), '--engine=docker'],
        )
        self.assertEqual(result.exit_code, 0)
        self._docker_client.api.build.assert_called_once()
        self._docker_client.images.push.assert_called_once_with(
            'custom-image', stream=True, decode=True)

    def test_kaniko_engine_is_not_supported(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)
        with self.assertWarnsRegex(DeprecationWarning,
                                   r'The --engine option is deprecated'):
            result = self.runner.invoke(
                self.cli,
                ['build', str(self._working_dir), '--engine=kaniko'],
            )
        self.assertEqual(result.exit_code, 1)
        self._docker_client.api.build.assert_not_called()
        self._docker_client.images.push.assert_not_called()

    def test_cloud_build_engine_is_not_supported(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        with self.assertWarnsRegex(DeprecationWarning,
                                   r'The --engine option is deprecated'):
            result = self.runner.invoke(
                self.cli,
                ['build',
                 str(self._working_dir), '--engine=cloudbuild'],
            )

        self.assertEqual(result.exit_code, 1)
        self._docker_client.api.build.assert_not_called()
        self._docker_client.images.push.assert_not_called()

    def test_docker_client_is_called_to_build_and_push_by_default(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)

        self._docker_client.api.build.assert_called_once_with(
            path=mock.ANY,
            dockerfile='Dockerfile',
            tag=mock.ANY,
            decode=True,
            platform='linux/amd64',
        )
        self._docker_client.images.push.assert_called_once_with(
            'custom-image', stream=True, decode=True)

    def test_docker_client_is_not_called_to_build_or_push(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir), '--no-build-image'],
        )
        self.assertEqual(result.exit_code, 0)

        self._docker_client.api.build.assert_not_called()
        self._docker_client.images.push.assert_not_called()

    def test_docker_client_is_called_to_build_but_skips_pushing(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir), '--no-push-image'],
        )
        self.assertEqual(result.exit_code, 0)

        self._docker_client.api.build.assert_called_once()
        self._docker_client.images.push.assert_not_called()

    @mock.patch('kfp.__version__', '1.2.3')
    def test_docker_file_is_created_correctly(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)
        self._docker_client.api.build.assert_called_once()
        self.assert_file_exists_and_contains(
            'Dockerfile',
            textwrap.dedent('''\
                # Generated by KFP.

                FROM python:3.7

                WORKDIR /usr/local/src/kfp/components
                COPY runtime-requirements.txt runtime-requirements.txt
                RUN pip install --no-cache-dir -r runtime-requirements.txt

                RUN pip install --no-cache-dir kfp==1.2.3
                COPY . .
                '''))

    @mock.patch('kfp.__version__', '1.2.3')
    def test_docker_file_is_created_correctly_with_one_url(self):
        component = _make_component(
            func_name='train',
            target_image='custom-image',
            pip_index_urls=['https://pypi.org/simple'])
        _write_components('components.py', component)

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)
        self._docker_client.api.build.assert_called_once()
        self.assert_file_exists_and_contains(
            'Dockerfile',
            textwrap.dedent('''\
                # Generated by KFP.

                FROM python:3.7

                WORKDIR /usr/local/src/kfp/components
                COPY runtime-requirements.txt runtime-requirements.txt
                RUN pip install --index-url https://pypi.org/simple --trusted-host https://pypi.org/simple --no-cache-dir -r runtime-requirements.txt

                RUN pip install --index-url https://pypi.org/simple --trusted-host https://pypi.org/simple --no-cache-dir kfp==1.2.3
                COPY . .
                '''))

    @mock.patch('kfp.__version__', '1.2.3')
    def test_docker_file_is_created_correctly_with_two_urls(self):
        component = _make_component(
            func_name='train',
            target_image='custom-image',
            pip_index_urls=[
                'https://pypi.org/simple', 'https://example.com/pypi/simple'
            ])
        _write_components('components.py', component)

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)
        self._docker_client.api.build.assert_called_once()
        self.assert_file_exists_and_contains(
            'Dockerfile',
            textwrap.dedent('''\
                # Generated by KFP.

                FROM python:3.7

                WORKDIR /usr/local/src/kfp/components
                COPY runtime-requirements.txt runtime-requirements.txt
                RUN pip install --index-url https://pypi.org/simple --trusted-host https://pypi.org/simple --extra-index-url https://example.com/pypi/simple --trusted-host https://example.com/pypi/simple --no-cache-dir -r runtime-requirements.txt

                RUN pip install --index-url https://pypi.org/simple --trusted-host https://pypi.org/simple --extra-index-url https://example.com/pypi/simple --trusted-host https://example.com/pypi/simple --no-cache-dir kfp==1.2.3
                COPY . .
                '''))

    def test_existing_dockerfile_is_unchanged_by_default(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)
        _write_file('Dockerfile', 'Existing Dockerfile contents')

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)
        self._docker_client.api.build.assert_called_once()
        self.assert_file_exists_and_contains('Dockerfile',
                                             'Existing Dockerfile contents')

    @mock.patch('kfp.__version__', '1.2.3')
    def test_existing_dockerfile_can_be_overwritten(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)
        _write_file('Dockerfile', 'Existing Dockerfile contents')

        result = self.runner.invoke(
            self.cli,
            ['build', str(self._working_dir), '--overwrite-dockerfile'],
        )
        self.assertEqual(result.exit_code, 0)
        self._docker_client.api.build.assert_called_once()
        self.assert_file_exists_and_contains(
            'Dockerfile',
            textwrap.dedent('''\
                # Generated by KFP.

                FROM python:3.7

                WORKDIR /usr/local/src/kfp/components
                COPY runtime-requirements.txt runtime-requirements.txt
                RUN pip install --no-cache-dir -r runtime-requirements.txt

                RUN pip install --no-cache-dir kfp==1.2.3
                COPY . .
                '''))

    def test_dockerfile_can_contain_custom_kfp_package(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)
        current_dir = os.path.dirname(os.path.abspath(__file__))
        package_dir = os.path.dirname(os.path.dirname(current_dir))

        # suppresses large stdout from subprocess that builds kfp package
        with mock.patch.object(
                subprocess,
                'run',
                new=functools.partial(subprocess.run, capture_output=True)):
            result = self.runner.invoke(
                self.cli,
                [
                    'build',
                    str(self._working_dir), f'--kfp-package-path={package_dir}'
                ],
            )
        self.assertEqual(result.exit_code, 0)
        self._docker_client.api.build.assert_called_once()
        self.assert_file_exists('Dockerfile')
        with open('Dockerfile', 'r') as f:
            contents = f.read()
        file_start = textwrap.dedent('''\
                # Generated by KFP.

                FROM python:3.7

                WORKDIR /usr/local/src/kfp/components
                COPY runtime-requirements.txt runtime-requirements.txt
                RUN pip install --no-cache-dir -r runtime-requirements.txt
                ''')
        self.assertTrue(contents.startswith(file_start))
        self.assertRegex(contents, 'RUN pip install --no-cache-dir kfp-*')


if __name__ == '__main__':
    unittest.main()



# Copyright 2019 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "run_pipeline_func_on_cluster",
    "run_pipeline_func_locally",
]

from typing import Callable, Mapping, Optional

from . import Client, LocalClient, dsl


def run_pipeline_func_on_cluster(
    pipeline_func: Callable,
    arguments: Mapping[str, str],
    run_name: str = None,
    experiment_name: str = None,
    kfp_client: Client = None,
    pipeline_conf: dsl.PipelineConf = None,
):
    """Runs pipeline on KFP-enabled Kubernetes cluster.

    This command compiles the pipeline function, creates or gets an experiment
    and submits the pipeline for execution.

    Feature stage:
    [Alpha](https://github.com/kubeflow/pipelines/blob/07328e5094ac2981d3059314cc848fbb71437a76/docs/release/feature-stages.md#alpha)

    Args:
      pipeline_func: A function that describes a pipeline by calling components
      and composing them into execution graph.
      arguments: Arguments to the pipeline function provided as a dict.
      run_name: Optional. Name of the run to be shown in the UI.
      experiment_name: Optional. Name of the experiment to add the run to.
      kfp_client: Optional. An instance of kfp.Client configured for the desired
        KFP cluster.
      pipeline_conf: Optional. kfp.dsl.PipelineConf instance. Can specify op
        transforms, image pull secrets and other pipeline-level configuration
        options.
    """
    kfp_client = kfp_client or Client()
    return kfp_client.create_run_from_pipeline_func(pipeline_func, arguments,
                                                    run_name, experiment_name,
                                                    pipeline_conf)


def run_pipeline_func_locally(
        pipeline_func: Callable,
        arguments: Mapping[str, str],
        local_client: Optional[LocalClient] = None,
        pipeline_root: Optional[str] = None,
        execution_mode: LocalClient.ExecutionMode = LocalClient.ExecutionMode(),
):
    """Runs a pipeline locally, either using Docker or in a local process.

    Feature stage:
    [Alpha](https://github.com/kubeflow/pipelines/blob/master/docs/release/feature-stages.md#alpha)

    In this alpha implementation, we support:
      * Control flow: Condition, ParallelFor
      * Data passing: InputValue, InputPath, OutputPath

    And we don't support:
      * Control flow: ExitHandler, Graph, SubGraph
      * ContainerOp with environment variables, init containers, sidecars, pvolumes
      * ResourceOp
      * VolumeOp
      * Caching

    Args:
      pipeline_func: A function that describes a pipeline by calling components
        and composing them into execution graph.
      arguments: Arguments to the pipeline function provided as a dict.
        reference to `kfp.client.create_run_from_pipeline_func`.
      local_client: Optional. An instance of kfp.LocalClient.
      pipeline_root: Optional. The root directory where the output artifact of component
        will be saved.
      execution_mode: Configuration to decide whether the client executes component
        in docker or in local process.
    """
    local_client = local_client or LocalClient(pipeline_root)
    return local_client.create_run_from_pipeline_func(
        pipeline_func, arguments, execution_mode=execution_mode)



# Copyright 2018 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from collections import namedtuple
from typing import Dict, List, Optional, Union

# TODO: Move this to a separate class
# For now, this identifies a condition with only "==" operator supported.
ConditionOperator = namedtuple('ConditionOperator',
                               'operator operand1 operand2')
PipelineParamTuple = namedtuple('PipelineParamTuple', 'name op pattern')


def sanitize_k8s_name(name, allow_capital_underscore=False):
    """Cleans and converts the names in the workflow.

    Args:
      name: original name,
      allow_capital_underscore: whether to allow capital letter and underscore in
        this name.

    Returns:
      A sanitized name.
    """
    if allow_capital_underscore:
        return re.sub('-+', '-', re.sub('[^-_0-9A-Za-z]+', '-',
                                        name)).lstrip('-').rstrip('-')
    else:
        return re.sub('-+', '-', re.sub('[^-0-9a-z]+', '-',
                                        name.lower())).lstrip('-').rstrip('-')


def match_serialized_pipelineparam(payload: str) -> List[PipelineParamTuple]:
    """Matches the supplied serialized pipelineparam.

    Args:
      payloads: The search space for the serialized pipelineparams.

    Returns:
      The matched pipeline params we found in the supplied payload.
    """
    matches = re.findall(r'{{pipelineparam:op=([\w\s_-]*);name=([\w\s_-]+)}}',
                         payload)
    param_tuples = []
    for match in matches:
        pattern = '{{pipelineparam:op=%s;name=%s}}' % (match[0], match[1])
        param_tuples.append(
            PipelineParamTuple(
                name=sanitize_k8s_name(match[1], True),
                op=sanitize_k8s_name(match[0]),
                pattern=pattern))
    return param_tuples


def _extract_pipelineparams(
        payloads: Union[str, List[str]]) -> List['PipelineParam']:
    """Extracts a list of PipelineParam instances from the payload string.

    Note: this function removes all duplicate matches.

    Args:
      payload: a string/a list of strings that contains serialized pipelineparams

    Return: List[]
    """
    if isinstance(payloads, str):
        payloads = [payloads]
    param_tuples = []
    for payload in payloads:
        param_tuples += match_serialized_pipelineparam(payload)
    pipeline_params = []
    for param_tuple in list(set(param_tuples)):
        pipeline_params.append(
            PipelineParam(
                param_tuple.name, param_tuple.op, pattern=param_tuple.pattern))
    return pipeline_params


def extract_pipelineparams_from_any(
    payload: Union['PipelineParam', str, list, tuple, dict]
) -> List['PipelineParam']:
    """Recursively extract PipelineParam instances or serialized string from
    any object or list of objects.

    Args:
      payload (str or k8_obj or list[str or k8_obj]): a string/a list of strings
        that contains serialized pipelineparams or a k8 definition object.

    Return: List[PipelineParam]
    """
    if not payload:
        return []

    # PipelineParam
    if isinstance(payload, PipelineParam):
        return [payload]

    # str
    if isinstance(payload, str):
        return list(set(_extract_pipelineparams(payload)))

    # list or tuple
    if isinstance(payload, list) or isinstance(payload, tuple):
        pipeline_params = []
        for item in payload:
            pipeline_params += extract_pipelineparams_from_any(item)
        return list(set(pipeline_params))

    # dict
    if isinstance(payload, dict):
        pipeline_params = []
        for key, value in payload.items():
            pipeline_params += extract_pipelineparams_from_any(key)
            pipeline_params += extract_pipelineparams_from_any(value)
        return list(set(pipeline_params))

    # k8s OpenAPI object
    if hasattr(payload, 'attribute_map') and isinstance(payload.attribute_map,
                                                        dict):
        pipeline_params = []
        for key in payload.attribute_map:
            pipeline_params += extract_pipelineparams_from_any(
                getattr(payload, key))

        return list(set(pipeline_params))

    # return empty list
    return []


class PipelineParam(object):
    """Representing a future value that is passed between pipeline components.

    A PipelineParam object can be used as a pipeline function argument so that it
    will be a pipeline parameter that shows up in ML Pipelines system UI. It can
    also represent an intermediate value passed between components.

    Args:
      name: name of the pipeline parameter.
      op_name: the name of the operation that produces the PipelineParam. None
        means it is not produced by any operator, so if None, either user
        constructs it directly (for providing an immediate value), or it is a
        pipeline function argument.
      value: The actual value of the PipelineParam. If provided, the PipelineParam
        is "resolved" immediately. For now, we support string only.
      param_type: the type of the PipelineParam.
      pattern: the serialized string regex pattern this pipeline parameter created
        from.

    Raises: ValueError in name or op_name contains invalid characters, or both
      op_name and value are set.
    """

    def __init__(self,
                 name: str,
                 op_name: Optional[str] = None,
                 value: Optional[str] = None,
                 param_type: Optional[Union[str, Dict]] = None,
                 pattern: Optional[str] = None):
        valid_name_regex = r'^[A-Za-z][A-Za-z0-9\s_-]*$'
        if not re.match(valid_name_regex, name):
            raise ValueError(
                'Only letters, numbers, spaces, "_", and "-" are allowed in name. '
                'Must begin with a letter. Got name: {}'.format(name))

        if op_name and value:
            raise ValueError('op_name and value cannot be both set.')

        self.name = name
        # ensure value is None even if empty string or empty list
        # so that serialization and unserialization remain consistent
        # (i.e. None => '' => None)
        self.op_name = op_name if op_name else None
        self.value = value
        self.param_type = param_type
        self.pattern = pattern or str(self)

    @property
    def full_name(self):
        """Unique name in the argo yaml for the PipelineParam."""
        if self.op_name:
            return self.op_name + '-' + self.name
        return self.name

    def __str__(self):
        """String representation.

        The string representation is a string identifier so we can mix
        the PipelineParam inline with other strings such as arguments.
        For example, we can support: ['echo %s' % param] as the
        container command and later a compiler can replace the
        placeholder "{{pipelineparam:op=%s;name=%s}}" with its own
        parameter identifier.
        """

        #This is deleted because if users specify default values to PipelineParam,
        # The compiler can not detect it as the value is not NULL.
        #if self.value:
        #  return str(self.value)

        op_name = self.op_name if self.op_name else ''
        return '{{pipelineparam:op=%s;name=%s}}' % (op_name, self.name)

    def __repr__(self):
        # return str({self.__class__.__name__: self.__dict__})
        # We make repr return the placeholder string so that if someone uses
        # str()-based serialization of complex objects containing `PipelineParam`,
        # it works properly.
        # (e.g. str([1, 2, 3, kfp.dsl.PipelineParam("aaa"), 4, 5, 6,]))
        return str(self)

    def to_struct(self):
        # Used by the json serializer. Outputs a JSON-serializable representation of
        # the object
        return str(self)

    def __eq__(self, other):
        return ConditionOperator('==', self, other)

    def __ne__(self, other):
        return ConditionOperator('!=', self, other)

    def __lt__(self, other):
        return ConditionOperator('<', self, other)

    def __le__(self, other):
        return ConditionOperator('<=', self, other)

    def __gt__(self, other):
        return ConditionOperator('>', self, other)

    def __ge__(self, other):
        return ConditionOperator('>=', self, other)

    def __hash__(self):
        return hash((self.op_name, self.name))

    def ignore_type(self):
        """ignore_type ignores the type information such that type checking
        would also pass."""
        self.param_type = None
        return self



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definitions of utils methods."""

import importlib
import os
import re
import sys
import types
from typing import List

_COMPONENT_NAME_PREFIX = 'comp-'
_EXECUTOR_LABEL_PREFIX = 'exec-'


def load_module(module_name: str, module_directory: str) -> types.ModuleType:
    """Dynamically imports the Python module with the given name and package
    path.

    E.g., Assuming there is a file called `my_module.py` under
    `/some/directory/my_module`, we can use::

        load_module('my_module', '/some/directory')

    to effectively `import mymodule`.

    Args:
        module_name: The name of the module.
        package_path: The package under which the specified module resides.
    """
    module_spec = importlib.util.spec_from_file_location(
        name=module_name,
        location=os.path.join(module_directory, f'{module_name}.py'))
    module = importlib.util.module_from_spec(module_spec)
    sys.modules[module_spec.name] = module
    sys.path.insert(0, str(module_directory))
    module_spec.loader.exec_module(module)
    return module


def maybe_rename_for_k8s(name: str) -> str:
    """Cleans and converts a name to be k8s compatible.

    Args:
      name: The original name.

    Returns:
      A sanitized name.
    """
    return re.sub('-+', '-', re.sub('[^-0-9a-z]+', '-',
                                    name.lower())).lstrip('-').rstrip('-')


def sanitize_input_name(name: str) -> str:
    """Sanitizes input name."""
    return re.sub('[^_0-9a-z]+', '_', name.lower()).lstrip('_').rstrip('_')


def sanitize_component_name(name: str) -> str:
    """Sanitizes component name."""
    return _COMPONENT_NAME_PREFIX + maybe_rename_for_k8s(name)


def sanitize_task_name(name: str) -> str:
    """Sanitizes task name."""
    return maybe_rename_for_k8s(name)


def sanitize_executor_label(label: str) -> str:
    """Sanitizes executor label."""
    return _EXECUTOR_LABEL_PREFIX + maybe_rename_for_k8s(label)


def make_name_unique_by_adding_index(
    name: str,
    collection: List[str],
    delimiter: str,
) -> str:
    """Makes a unique name by adding index.

    The index starts from 2 and increase by 1 until we find a unique name.

    Args:
        name: The original name.
        collection: The collection of existing names.
        delimiter: The delimiter to connect the original name and an index.

    Returns:
        A unique name composed of name+delimiter+next index
    """
    unique_name = name
    if unique_name in collection:
        for i in range(2, sys.maxsize**10):
            unique_name = name + delimiter + str(i)
            if unique_name not in collection:
                break
    return unique_name


def validate_pipeline_name(name: str) -> None:
    """Validate pipeline name.

    A valid pipeline name should match ^[a-z0-9][a-z0-9-]{0,127}$.

    Args:
        name: The pipeline name.

    Raises:
        ValueError if the pipeline name doesn't conform to the regular expression.
    """
    pattern = re.compile(r'^[a-z0-9][a-z0-9-]{0,127}$')
    if not pattern.match(name):
        raise ValueError(
            'Invalid pipeline name: %s.\n'
            'Please specify a pipeline name that matches the regular '
            'expression "^[a-z0-9][a-z0-9-]{0,127}$" using '
            '`dsl.pipeline(name=...)` decorator.' % name)



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from absl.testing import parameterized
from kfp.compiler import compiler_utils
from kfp.components import pipeline_channel


class TestAdditionalInputNameForPipelineChannel(parameterized.TestCase):

    @parameterized.parameters(
        {
            'channel':
                pipeline_channel.create_pipeline_channel(
                    name='output1', task_name='task1', channel_type='String'),
            'expected':
                'pipelinechannel--task1-output1',
        },
        {
            'channel':
                pipeline_channel.create_pipeline_channel(
                    name='output1',
                    task_name='task1',
                    channel_type='system.Artifact@0.0.1',
                ),
            'expected':
                'pipelinechannel--task1-output1',
        },
        {
            'channel':
                pipeline_channel.create_pipeline_channel(
                    name='param1', channel_type='String'),
            'expected':
                'pipelinechannel--param1',
        },
    )
    def test_additional_input_name_for_pipeline_channel(self, channel,
                                                        expected):
        self.assertEqual(
            expected,
            compiler_utils.additional_input_name_for_pipeline_channel(channel))


if __name__ == '__main__':
    unittest.main()



# Copyright 2019 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import tempfile


class GCSHelper(object):
    """GCSHelper manages the connection with the GCS storage."""

    @staticmethod
    def get_blob_from_gcs_uri(gcs_path):
        """
    Args:
      gcs_path (str) : gcs blob path
    Returns:
      gcs_blob: gcs blob object(https://github.com/googleapis/google-cloud-python/blob/5c9bb42cb3c9250131cfeef6e0bafe8f4b7c139f/storage/google/cloud/storage/blob.py#L105)
    """
        from google.cloud import storage
        pure_path = pathlib.PurePath(gcs_path)
        gcs_bucket = pure_path.parts[1]
        gcs_blob = '/'.join(pure_path.parts[2:])
        client = storage.Client()
        bucket = client.get_bucket(gcs_bucket)
        blob = bucket.blob(gcs_blob)
        return blob

    @staticmethod
    def upload_gcs_file(local_path, gcs_path):
        """
    Args:
      local_path (str): local file path
      gcs_path (str) : gcs blob path
    """
        blob = GCSHelper.get_blob_from_gcs_uri(gcs_path)
        blob.upload_from_filename(local_path)

    @staticmethod
    def write_to_gcs_path(path: str, content: str) -> None:
        """Writes serialized content to a GCS location.

        Args:
          path: GCS path to write to.
          content: The content to be written.
        """
        fd, temp_path = tempfile.mkstemp()
        try:
            with os.fdopen(fd, 'w') as tmp:
                tmp.write(content)

            if not GCSHelper.get_blob_from_gcs_uri(path):
                pure_path = pathlib.PurePath(path)
                gcs_bucket = pure_path.parts[1]
                GCSHelper.create_gcs_bucket_if_not_exist(gcs_bucket)

            GCSHelper.upload_gcs_file(temp_path, path)
        finally:
            os.remove(temp_path)

    @staticmethod
    def remove_gcs_blob(gcs_path):
        """
    Args:
      gcs_path (str) : gcs blob path
    """
        blob = GCSHelper.get_blob_from_gcs_uri(gcs_path)
        blob.delete()

    @staticmethod
    def download_gcs_blob(local_path, gcs_path):
        """
    Args:
      local_path (str): local file path
      gcs_path (str) : gcs blob path
    """
        blob = GCSHelper.get_blob_from_gcs_uri(gcs_path)
        blob.download_to_filename(local_path)

    @staticmethod
    def read_from_gcs_path(gcs_path: str) -> str:
        """Reads the content of a file hosted on GCS."""
        fd, temp_path = tempfile.mkstemp()
        try:
            GCSHelper.download_gcs_blob(temp_path, gcs_path)
            with os.fdopen(fd, 'r') as tmp:
                result = tmp.read()
        finally:
            os.remove(temp_path)
        return result

    @staticmethod
    def create_gcs_bucket_if_not_exist(gcs_bucket):
        """
    Args:
      gcs_bucket (str) : gcs bucket name
    """
        from google.cloud import storage
        from google.cloud.exceptions import NotFound
        client = storage.Client()
        try:
            client.get_bucket(gcs_bucket)
        except NotFound:
            client.create_bucket(gcs_bucket)



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for input/output type annotations in KFP SDK.

These are only compatible with v2 Pipelines.
"""

import re
from typing import List, Type, TypeVar, Union

from kfp.components.types import artifact_types
from kfp.components.types import type_annotations
from kfp.components.types import type_utils


class OutputPath:
    """Type annotation used in component definitions for indicating a parameter
    is a path to an output. The path parameter typed with this annotation can
    be treated as a locally accessible filepath within the component body.

    The argument typed with this annotation is provided at runtime by the executing backend and does not need to be passed as an input by the pipeline author (see example).


    Args:
        type: The type of the value written to the output path.

    Example:
      ::

        @dsl.component
        def create_parameter(
                message: str,
                output_parameter_path: OutputPath(str),
        ):
            with open(output_parameter_path, 'w') as f:
                f.write(message)


        @dsl.component
        def consume_parameter(message: str):
            print(message)


        @dsl.pipeline(name='my-pipeline', pipeline_root='gs://my-bucket')
        def my_pipeline(message: str = 'default message'):
            create_param_op = create_parameter(message=message)
            consume_parameter(message=create_param_op.outputs['output_parameter_path'])
    """

    def __init__(self, type=None):
        self.type = construct_type_for_inputpath_or_outputpath(type)

    def __eq__(self, other):
        return isinstance(other, OutputPath) and self.type == other.type


class InputPath:
    """Type annotation used in component definitions for indicating a parameter
    is a path to an input.

    Example:
      ::

        @dsl.component
        def create_dataset(dataset_path: OutputPath('Dataset'),):
            import json
            dataset = {'my_dataset': [[1, 2, 3], [4, 5, 6]]}
            with open(dataset_path, 'w') as f:
                json.dump(dataset, f)


        @dsl.component
        def consume_dataset(dataset: InputPath('Dataset')):
            print(dataset)


        @dsl.pipeline(name='my-pipeline', pipeline_root='gs://my-bucket')
        def my_pipeline():
            create_dataset_op = create_dataset()
            consume_dataset(dataset=create_dataset_op.outputs['dataset_path'])
    """

    def __init__(self, type=None):
        self.type = construct_type_for_inputpath_or_outputpath(type)

    def __eq__(self, other):
        return isinstance(other, InputPath) and self.type == other.type


def construct_type_for_inputpath_or_outputpath(
        type_: Union[str, Type, None]) -> Union[str, None]:
    if type_annotations.is_artifact_class(type_):
        return type_utils.create_bundled_artifact_type(type_.schema_title,
                                                       type_.schema_version)
    elif isinstance(
            type_,
            str) and type_.lower() in type_utils._ARTIFACT_CLASSES_MAPPING:
        # v1 artifact backward compat, e.g. dsl.OutputPath('Dataset')
        return type_utils.create_bundled_artifact_type(
            type_utils._ARTIFACT_CLASSES_MAPPING[type_.lower()].schema_title)
    elif type_utils.get_parameter_type(type_):
        return type_
    else:
        # v1 unknown type dsl.OutputPath('MyCustomType')
        return type_utils.create_bundled_artifact_type(
            artifact_types.Artifact.schema_title)


class InputAnnotation:
    """Marker type for input artifacts."""


class OutputAnnotation:
    """Marker type for output artifacts."""


def is_Input_Output_artifact_annotation(typ) -> bool:
    if not hasattr(typ, '__metadata__'):
        return False

    if typ.__metadata__[0] not in [InputAnnotation, OutputAnnotation]:
        return False

    return True


def is_input_artifact(typ) -> bool:
    """Returns True if typ is of type Input[T]."""
    if not is_Input_Output_artifact_annotation(typ):
        return False

    return typ.__metadata__[0] == InputAnnotation


def is_output_artifact(typ) -> bool:
    """Returns True if typ is of type Output[T]."""
    if not is_Input_Output_artifact_annotation(typ):
        return False

    return typ.__metadata__[0] == OutputAnnotation


def get_io_artifact_class(typ):
    from kfp.dsl import Input
    from kfp.dsl import Output
    if not is_Input_Output_artifact_annotation(typ):
        return None
    if typ == Input or typ == Output:
        return None

    # extract inner type from list of artifacts
    inner = typ.__args__[0]
    if hasattr(inner, '__origin__') and inner.__origin__ == list:
        return inner.__args__[0]

    return inner


def get_io_artifact_annotation(typ):
    if not is_Input_Output_artifact_annotation(typ):
        return None

    return typ.__metadata__[0]


T = TypeVar('T')


def maybe_strip_optional_from_annotation(annotation: T) -> T:
    """Strips 'Optional' from 'Optional[<type>]' if applicable.

    For example::
      Optional[str] -> str
      str -> str
      List[int] -> List[int]

    Args:
      annotation: The original type annotation which may or may not has
        `Optional`.

    Returns:
      The type inside Optional[] if Optional exists, otherwise the original type.
    """
    if getattr(annotation, '__origin__',
               None) is Union and annotation.__args__[1] is type(None):
        return annotation.__args__[0]
    return annotation


def maybe_strip_optional_from_annotation_string(annotation: str) -> str:
    if annotation.startswith('Optional[') and annotation.endswith(']'):
        return annotation.lstrip('Optional[').rstrip(']')
    return annotation


def get_short_type_name(type_name: str) -> str:
    """Extracts the short form type name.

    This method is used for looking up serializer for a given type.

    For example::
      typing.List -> List
      typing.List[int] -> List
      typing.Dict[str, str] -> Dict
      List -> List
      str -> str

    Args:
      type_name: The original type name.

    Returns:
      The short form type name or the original name if pattern doesn't match.
    """
    match = re.match('(typing\.)?(?P<type>\w+)(?:\[.+\])?', type_name)
    return match['type'] if match else type_name


def is_artifact_class(artifact_class_or_instance: Type) -> bool:
    # we do not yet support non-pre-registered custom artifact types with instance_schema attribute
    return hasattr(artifact_class_or_instance, 'schema_title') and hasattr(
        artifact_class_or_instance, 'schema_version')


def is_list_of_artifacts(
    type_var: Union[Type[List[artifact_types.Artifact]],
                    Type[artifact_types.Artifact]]
) -> bool:
    # the type annotation for this function's `type_var` parameter may not actually be a subclass of the KFP SDK's Artifact class for custom artifact types
    is_list_or_list_generic = getattr(type_var, '__origin__', None) == list
    # in >= python3.9, List wont have .__args__ if it's used as `-> List` with no inner type argument
    contains_artifact = hasattr(
        type_var, '__args__') and type_annotations.is_artifact_class(
            type_var.__args__[0])
    return is_list_or_list_generic and contains_artifact



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.dsl.component_spec."""

from absl.testing import parameterized

from kfp.deprecated.components import _structures as structures
from kfp.deprecated.dsl import _pipeline_param
from kfp.deprecated.dsl import component_spec as dsl_component_spec
from kfp.pipeline_spec import pipeline_spec_pb2

from google.protobuf import json_format


class ComponentSpecTest(parameterized.TestCase):

    TEST_PIPELINE_PARAMS = [
        _pipeline_param.PipelineParam(
            name='output1', param_type='Dataset', op_name='op-1'),
        _pipeline_param.PipelineParam(
            name='output2', param_type='Integer', op_name='op-2'),
        _pipeline_param.PipelineParam(
            name='output3', param_type='Model', op_name='op-3'),
        _pipeline_param.PipelineParam(
            name='output4', param_type='Double', op_name='op-4'),
        _pipeline_param.PipelineParam(
            name='arg_input', param_type='String', op_name=None),
    ]

    def setUp(self):
        self.maxDiff = None

    def test_build_component_spec_from_structure(self):
        structure_component_spec = structures.ComponentSpec(
            name='component1',
            description='component1 desc',
            inputs=[
                structures.InputSpec(
                    name='input1', description='input1 desc', type='Dataset'),
                structures.InputSpec(
                    name='input2', description='input2 desc', type='String'),
                structures.InputSpec(
                    name='input3', description='input3 desc', type='Integer'),
                structures.InputSpec(
                    name='input4', description='optional inputs',
                    optional=True),
            ],
            outputs=[
                structures.OutputSpec(
                    name='output1', description='output1 desc', type='Model')
            ])
        expected_dict = {
            'inputDefinitions': {
                'artifacts': {
                    'input1': {
                        'artifactType': {
                            'schemaTitle': 'system.Dataset',
                            'schemaVersion': '0.0.1'
                        }
                    }
                },
                'parameters': {
                    'input2': {
                        'parameterType': 'STRING'
                    },
                    'input3': {
                        'parameterType': 'NUMBER_INTEGER'
                    }
                }
            },
            'outputDefinitions': {
                'artifacts': {
                    'output1': {
                        'artifactType': {
                            'schemaTitle': 'system.Model',
                            'schemaVersion': '0.0.1'
                        }
                    }
                }
            },
            'executorLabel': 'exec-component1'
        }
        expected_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(expected_dict, expected_spec)

        component_spec = (
            dsl_component_spec.build_component_spec_from_structure(
                component_spec=structure_component_spec,
                executor_label='exec-component1',
                actual_inputs=['input1', 'input2', 'input3'],
            ))

        self.assertEqual(expected_spec, component_spec)

    @parameterized.parameters(
        {
            'is_root_component': True,
            'expected_result': {
                'inputDefinitions': {
                    'artifacts': {
                        'input1': {
                            'artifactType': {
                                'schemaTitle': 'system.Dataset',
                                'schemaVersion': '0.0.1'
                            }
                        }
                    },
                    'parameters': {
                        'input2': {
                            'parameterType': 'NUMBER_INTEGER'
                        },
                        'input3': {
                            'parameterType': 'STRING'
                        },
                        'input4': {
                            'parameterType': 'NUMBER_DOUBLE'
                        }
                    }
                }
            }
        },
        {
            'is_root_component': False,
            'expected_result': {
                'inputDefinitions': {
                    'artifacts': {
                        'pipelineparam--input1': {
                            'artifactType': {
                                'schemaTitle': 'system.Dataset',
                                'schemaVersion': '0.0.1'
                            }
                        }
                    },
                    'parameters': {
                        'pipelineparam--input2': {
                            'parameterType': 'NUMBER_INTEGER'
                        },
                        'pipelineparam--input3': {
                            'parameterType': 'STRING'
                        },
                        'pipelineparam--input4': {
                            'parameterType': 'NUMBER_DOUBLE'
                        }
                    }
                }
            }
        },
    )
    def test_build_component_inputs_spec(self, is_root_component,
                                         expected_result):
        pipeline_params = [
            _pipeline_param.PipelineParam(name='input1', param_type='Dataset'),
            _pipeline_param.PipelineParam(name='input2', param_type='Integer'),
            _pipeline_param.PipelineParam(name='input3', param_type='String'),
            _pipeline_param.PipelineParam(name='input4', param_type='Float'),
        ]
        expected_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(expected_result, expected_spec)

        component_spec = pipeline_spec_pb2.ComponentSpec()
        dsl_component_spec.build_component_inputs_spec(component_spec,
                                                       pipeline_params,
                                                       is_root_component)

        self.assertEqual(expected_spec, component_spec)

    def test_build_component_outputs_spec(self):
        pipeline_params = [
            _pipeline_param.PipelineParam(name='output1', param_type='Dataset'),
            _pipeline_param.PipelineParam(name='output2', param_type='Integer'),
            _pipeline_param.PipelineParam(name='output3', param_type='String'),
            _pipeline_param.PipelineParam(name='output4', param_type='Float'),
        ]
        expected_dict = {
            'outputDefinitions': {
                'artifacts': {
                    'output1': {
                        'artifactType': {
                            'schemaTitle': 'system.Dataset',
                            'schemaVersion': '0.0.1'
                        }
                    }
                },
                'parameters': {
                    'output2': {
                        'parameterType': 'NUMBER_INTEGER'
                    },
                    'output3': {
                        'parameterType': 'STRING'
                    },
                    'output4': {
                        'parameterType': 'NUMBER_DOUBLE'
                    }
                }
            }
        }
        expected_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(expected_dict, expected_spec)

        component_spec = pipeline_spec_pb2.ComponentSpec()
        dsl_component_spec.build_component_outputs_spec(component_spec,
                                                        pipeline_params)

        self.assertEqual(expected_spec, component_spec)

    @parameterized.parameters(
        {
            'is_parent_component_root': True,
            'expected_result': {
                'inputs': {
                    'artifacts': {
                        'pipelineparam--op-1-output1': {
                            'taskOutputArtifact': {
                                'producerTask': 'op-1',
                                'outputArtifactKey': 'output1'
                            }
                        },
                        'pipelineparam--op-3-output3': {
                            'componentInputArtifact': 'op-3-output3'
                        }
                    },
                    'parameters': {
                        'pipelineparam--op-2-output2': {
                            'taskOutputParameter': {
                                'producerTask': 'op-2',
                                'outputParameterKey': 'output2'
                            }
                        },
                        'pipelineparam--op-4-output4': {
                            'componentInputParameter': 'op-4-output4'
                        },
                        'pipelineparam--arg_input': {
                            'componentInputParameter': 'arg_input'
                        }
                    }
                }
            }
        },
        {
            'is_parent_component_root': False,
            'expected_result': {
                'inputs': {
                    'artifacts': {
                        'pipelineparam--op-1-output1': {
                            'taskOutputArtifact': {
                                'producerTask': 'op-1',
                                'outputArtifactKey': 'output1'
                            }
                        },
                        'pipelineparam--op-3-output3': {
                            'componentInputArtifact':
                                'pipelineparam--op-3-output3'
                        }
                    },
                    'parameters': {
                        'pipelineparam--op-2-output2': {
                            'taskOutputParameter': {
                                'producerTask': 'op-2',
                                'outputParameterKey': 'output2'
                            }
                        },
                        'pipelineparam--op-4-output4': {
                            'componentInputParameter':
                                'pipelineparam--op-4-output4'
                        },
                        'pipelineparam--arg_input': {
                            'componentInputParameter':
                                'pipelineparam--arg_input'
                        }
                    }
                }
            }
        },
    )
    def test_build_task_inputs_spec(self, is_parent_component_root,
                                    expected_result):
        pipeline_params = self.TEST_PIPELINE_PARAMS
        tasks_in_current_dag = ['op-1', 'op-2']
        expected_spec = pipeline_spec_pb2.PipelineTaskSpec()
        json_format.ParseDict(expected_result, expected_spec)

        task_spec = pipeline_spec_pb2.PipelineTaskSpec()
        dsl_component_spec.build_task_inputs_spec(task_spec, pipeline_params,
                                                  tasks_in_current_dag,
                                                  is_parent_component_root)

        self.assertEqual(expected_spec, task_spec)

    @parameterized.parameters(
        {
            'original_task_spec': {},
            'parent_component_inputs': {},
            'tasks_in_current_dag': [],
            'input_parameters_in_current_dag': [],
            'input_artifacts_in_current_dag': [],
            'expected_result': {},
        },
        { # Depending on tasks & inputs within the current DAG.
            'original_task_spec': {
                'inputs': {
                    'artifacts': {
                        'pipelineparam--op-1-output1': {
                            'taskOutputArtifact': {
                                'producerTask': 'op-1',
                                'outputArtifactKey': 'output1'
                            }
                        },
                        'artifact1': {
                          'componentInputArtifact': 'artifact1'
                        },
                    },
                    'parameters': {
                        'pipelineparam--op-2-output2': {
                            'taskOutputParameter': {
                                'producerTask': 'op-2',
                                'outputParameterKey': 'output2'
                            }
                        },
                        'param1': {
                          'componentInputParameter': 'param1'
                        },
                    }
                }
            },
            'parent_component_inputs': {
              'artifacts': {
                'artifact1': {
                  'artifactType': {
                    'instanceSchema': 'dummy_schema'
                  }
                },
              },
              'parameters': {
                'param1': {
                  'parameterType': 'STRING'
                },
              }
            },
            'tasks_in_current_dag': ['op-1', 'op-2'],
            'input_parameters_in_current_dag': ['param1'],
            'input_artifacts_in_current_dag': ['artifact1'],
            'expected_result': {
                'inputs': {
                    'artifacts': {
                        'pipelineparam--op-1-output1': {
                            'taskOutputArtifact': {
                                'producerTask': 'op-1',
                                'outputArtifactKey': 'output1'
                            }
                        },
                        'artifact1': {
                          'componentInputArtifact': 'artifact1'
                        },
                    },
                    'parameters': {
                        'pipelineparam--op-2-output2': {
                            'taskOutputParameter': {
                                'producerTask': 'op-2',
                                'outputParameterKey': 'output2'
                            }
                        },
                        'param1': {
                          'componentInputParameter': 'param1'
                        },
                    }
                }
            },
        },
        { # Depending on tasks and inputs not available in the current DAG.
            'original_task_spec': {
                'inputs': {
                    'artifacts': {
                        'pipelineparam--op-1-output1': {
                            'taskOutputArtifact': {
                                'producerTask': 'op-1',
                                'outputArtifactKey': 'output1'
                            }
                        },
                        'artifact1': {
                          'componentInputArtifact': 'artifact1'
                        },
                    },
                    'parameters': {
                        'pipelineparam--op-2-output2': {
                            'taskOutputParameter': {
                                'producerTask': 'op-2',
                                'outputParameterKey': 'output2'
                            }
                        },
                        'param1': {
                          'componentInputParameter': 'param1'
                        },
                    }
                }
            },
            'parent_component_inputs': {
                'artifacts': {
                    'pipelineparam--op-1-output1': {
                        'artifactType': {
                            'instanceSchema': 'dummy_schema'
                        }
                    },
                    'pipelineparam--artifact1': {
                      'artifactType': {
                        'instanceSchema': 'dummy_schema'
                      }
                    },
                },
                'parameters': {
                  'pipelineparam--op-2-output2' : {
                    'parameterType': 'NUMBER_INTEGER'
                  },
                  'pipelineparam--param1': {
                    'parameterType': 'STRING'
                  },
                }
            },
            'tasks_in_current_dag': ['op-3'],
            'input_parameters_in_current_dag': ['pipelineparam--op-2-output2', 'pipelineparam--param1'],
            'input_artifacts_in_current_dag': ['pipelineparam--op-1-output1', 'pipelineparam--artifact1'],
            'expected_result': {
                'inputs': {
                    'artifacts': {
                        'pipelineparam--op-1-output1': {
                            'componentInputArtifact':
                                'pipelineparam--op-1-output1'
                        },
                        'artifact1': {
                          'componentInputArtifact': 'pipelineparam--artifact1'
                        },
                    },
                    'parameters': {
                        'pipelineparam--op-2-output2': {
                            'componentInputParameter':
                                'pipelineparam--op-2-output2'
                        },
                        'param1': {
                          'componentInputParameter': 'pipelineparam--param1'
                        },
                    }
                }
            },
        },
    )
    def test_update_task_inputs_spec(self, original_task_spec,
                                     parent_component_inputs,
                                     tasks_in_current_dag,
                                     input_parameters_in_current_dag,
                                     input_artifacts_in_current_dag,
                                     expected_result):
        pipeline_params = self.TEST_PIPELINE_PARAMS

        expected_spec = pipeline_spec_pb2.PipelineTaskSpec()
        json_format.ParseDict(expected_result, expected_spec)

        task_spec = pipeline_spec_pb2.PipelineTaskSpec()
        json_format.ParseDict(original_task_spec, task_spec)
        parent_component_inputs_spec = pipeline_spec_pb2.ComponentInputsSpec()
        json_format.ParseDict(parent_component_inputs,
                              parent_component_inputs_spec)
        dsl_component_spec.update_task_inputs_spec(
            task_spec, parent_component_inputs_spec, pipeline_params,
            tasks_in_current_dag, input_parameters_in_current_dag,
            input_artifacts_in_current_dag)

        self.assertEqual(expected_spec, task_spec)

    def test_pop_input_from_component_spec(self):
        component_spec = pipeline_spec_pb2.ComponentSpec(
            executor_label='exec-component1')

        component_spec.input_definitions.artifacts[
            'input1'].artifact_type.schema_title = 'system.Dataset'
        component_spec.input_definitions.parameters[
            'input2'].parameter_type = pipeline_spec_pb2.ParameterType.STRING
        component_spec.input_definitions.parameters[
            'input3'].parameter_type = pipeline_spec_pb2.ParameterType.NUMBER_DOUBLE

        # pop an artifact, and there're other inputs left
        dsl_component_spec.pop_input_from_component_spec(
            component_spec, 'input1')
        expected_dict = {
            'inputDefinitions': {
                'parameters': {
                    'input2': {
                        'parameterType': 'STRING'
                    },
                    'input3': {
                        'parameterType': 'NUMBER_DOUBLE'
                    }
                }
            },
            'executorLabel': 'exec-component1'
        }
        expected_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(expected_dict, expected_spec)
        self.assertEqual(expected_spec, component_spec)

        # pop an parameter, and there're other inputs left
        dsl_component_spec.pop_input_from_component_spec(
            component_spec, 'input2')
        expected_dict = {
            'inputDefinitions': {
                'parameters': {
                    'input3': {
                        'parameterType': 'NUMBER_DOUBLE'
                    }
                }
            },
            'executorLabel': 'exec-component1'
        }
        expected_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(expected_dict, expected_spec)
        self.assertEqual(expected_spec, component_spec)

        # pop the last input, expect no inputDefinitions
        dsl_component_spec.pop_input_from_component_spec(
            component_spec, 'input3')
        expected_dict = {'executorLabel': 'exec-component1'}
        expected_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(expected_dict, expected_spec)
        self.assertEqual(expected_spec, component_spec)

        # pop an input that doesn't exist, expect no-op.
        dsl_component_spec.pop_input_from_component_spec(
            component_spec, 'input4')
        self.assertEqual(expected_spec, component_spec)

    def test_pop_input_from_task_spec(self):
        task_spec = pipeline_spec_pb2.PipelineTaskSpec()
        task_spec.component_ref.name = 'comp-component1'
        task_spec.inputs.artifacts[
            'input1'].task_output_artifact.producer_task = 'op-1'
        task_spec.inputs.artifacts[
            'input1'].task_output_artifact.output_artifact_key = 'output1'
        task_spec.inputs.parameters[
            'input2'].task_output_parameter.producer_task = 'op-2'
        task_spec.inputs.parameters[
            'input2'].task_output_parameter.output_parameter_key = 'output2'
        task_spec.inputs.parameters[
            'input3'].component_input_parameter = 'op3-output3'

        # pop an parameter, and there're other inputs left
        dsl_component_spec.pop_input_from_task_spec(task_spec, 'input3')
        expected_dict = {
            'inputs': {
                'artifacts': {
                    'input1': {
                        'taskOutputArtifact': {
                            'producerTask': 'op-1',
                            'outputArtifactKey': 'output1'
                        }
                    }
                },
                'parameters': {
                    'input2': {
                        'taskOutputParameter': {
                            'producerTask': 'op-2',
                            'outputParameterKey': 'output2'
                        }
                    }
                }
            },
            'component_ref': {
                'name': 'comp-component1'
            }
        }
        expected_spec = pipeline_spec_pb2.PipelineTaskSpec()
        json_format.ParseDict(expected_dict, expected_spec)
        self.assertEqual(expected_spec, task_spec)

        # pop an artifact, and there're other inputs left
        dsl_component_spec.pop_input_from_task_spec(task_spec, 'input1')
        expected_dict = {
            'inputs': {
                'parameters': {
                    'input2': {
                        'taskOutputParameter': {
                            'producerTask': 'op-2',
                            'outputParameterKey': 'output2'
                        }
                    }
                }
            },
            'component_ref': {
                'name': 'comp-component1'
            }
        }
        expected_spec = pipeline_spec_pb2.PipelineTaskSpec()
        json_format.ParseDict(expected_dict, expected_spec)
        self.assertEqual(expected_spec, task_spec)

        # pop the last input, expect no inputDefinitions
        dsl_component_spec.pop_input_from_task_spec(task_spec, 'input2')
        expected_dict = {'component_ref': {'name': 'comp-component1'}}
        expected_spec = pipeline_spec_pb2.PipelineTaskSpec()
        json_format.ParseDict(expected_dict, expected_spec)
        self.assertEqual(expected_spec, task_spec)

        # pop an input that doesn't exist, expect no-op.
        dsl_component_spec.pop_input_from_task_spec(task_spec, 'input4')
        self.assertEqual(expected_spec, task_spec)

    def test_additional_input_name_for_pipelineparam(self):
        self.assertEqual(
            'pipelineparam--op1-param1',
            dsl_component_spec.additional_input_name_for_pipelineparam(
                _pipeline_param.PipelineParam(name='param1', op_name='op1')))
        self.assertEqual(
            'pipelineparam--param2',
            dsl_component_spec.additional_input_name_for_pipelineparam(
                _pipeline_param.PipelineParam(name='param2')))
        self.assertEqual(
            'pipelineparam--param3',
            dsl_component_spec.additional_input_name_for_pipelineparam(
                'param3'))


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.components.types.type_annotations."""

from typing import Any, Dict, List, Optional
import unittest

from absl.testing import parameterized
from kfp.components.types import artifact_types
from kfp.components.types import type_annotations
from kfp.components.types.artifact_types import Model
from kfp.components.types.type_annotations import InputAnnotation
from kfp.components.types.type_annotations import InputPath
from kfp.components.types.type_annotations import OutputAnnotation
from kfp.components.types.type_annotations import OutputPath
from kfp.dsl import Input
from kfp.dsl import Output


class AnnotationsTest(parameterized.TestCase):

    @parameterized.parameters([
        Input[Model],
        Output[Model],
        Output[List[Model]],
        Output['MyArtifact'],
    ])
    def test_is_artifact_annotation(self, annotation):
        self.assertTrue(
            type_annotations.is_Input_Output_artifact_annotation(annotation))

    @parameterized.parameters([
        Model,
        int,
        'Dataset',
        List[str],
        Optional[str],
    ])
    def test_is_not_artifact_annotation(self, annotation):
        self.assertFalse(
            type_annotations.is_Input_Output_artifact_annotation(annotation))

    @parameterized.parameters([
        Input[Model],
        Input,
    ])
    def test_is_input_artifact(self, annotation):
        self.assertTrue(type_annotations.is_input_artifact(annotation))

    @parameterized.parameters([
        Output[Model],
        Output,
    ])
    def test_is_not_input_artifact(self, annotation):
        self.assertFalse(type_annotations.is_input_artifact(annotation))

    @parameterized.parameters([
        Output[Model],
        Output[List[Model]],
    ])
    def test_is_output_artifact(self, annotation):
        self.assertTrue(type_annotations.is_output_artifact(annotation))

    @parameterized.parameters([
        Input[Model],
        Input[List[Model]],
        Input,
    ])
    def test_is_not_output_artifact(self, annotation):
        self.assertFalse(type_annotations.is_output_artifact(annotation))

    def test_get_io_artifact_class(self):
        self.assertEqual(
            type_annotations.get_io_artifact_class(Output[Model]), Model)
        self.assertEqual(
            type_annotations.get_io_artifact_class(Output[List[Model]]), Model)
        self.assertEqual(
            type_annotations.get_io_artifact_class(Input[List[Model]]), Model)

        self.assertEqual(type_annotations.get_io_artifact_class(Input), None)
        self.assertEqual(type_annotations.get_io_artifact_class(Output), None)
        self.assertEqual(type_annotations.get_io_artifact_class(Model), None)
        self.assertEqual(type_annotations.get_io_artifact_class(str), None)

    def test_get_io_artifact_annotation(self):
        self.assertEqual(
            type_annotations.get_io_artifact_annotation(Output[Model]),
            OutputAnnotation)
        self.assertEqual(
            type_annotations.get_io_artifact_annotation(Output[List[Model]]),
            OutputAnnotation)
        self.assertEqual(
            type_annotations.get_io_artifact_annotation(Input[Model]),
            InputAnnotation)
        self.assertEqual(
            type_annotations.get_io_artifact_annotation(Input[List[Model]]),
            InputAnnotation)
        self.assertEqual(
            type_annotations.get_io_artifact_annotation(Input), InputAnnotation)
        self.assertEqual(
            type_annotations.get_io_artifact_annotation(Output),
            OutputAnnotation)

        self.assertEqual(
            type_annotations.get_io_artifact_annotation(Model), None)
        self.assertEqual(type_annotations.get_io_artifact_annotation(str), None)

    @parameterized.parameters(
        {
            'original_annotation': str,
            'expected_annotation': str,
        },
        {
            'original_annotation': 'MyCustomType',
            'expected_annotation': 'MyCustomType',
        },
        {
            'original_annotation': List[int],
            'expected_annotation': List[int],
        },
        {
            'original_annotation': Optional[str],
            'expected_annotation': str,
        },
        {
            'original_annotation': Optional[Dict[str, float]],
            'expected_annotation': Dict[str, float],
        },
        {
            'original_annotation': Optional[List[Dict[str, Any]]],
            'expected_annotation': List[Dict[str, Any]],
        },
        {
            'original_annotation': Input[Model],
            'expected_annotation': Input[Model],
        },
        {
            'original_annotation': InputPath('Model'),
            'expected_annotation': InputPath('Model'),
        },
        {
            'original_annotation': OutputPath(Model),
            'expected_annotation': OutputPath(Model),
        },
    )
    def test_maybe_strip_optional_from_annotation(self, original_annotation,
                                                  expected_annotation):
        self.assertEqual(
            expected_annotation,
            type_annotations.maybe_strip_optional_from_annotation(
                original_annotation))

    @parameterized.parameters(
        {
            'original_type_name': 'str',
            'expected_type_name': 'str',
        },
        {
            'original_type_name': 'typing.List[int]',
            'expected_type_name': 'List',
        },
        {
            'original_type_name': 'List[int]',
            'expected_type_name': 'List',
        },
        {
            'original_type_name': 'Dict[str, str]',
            'expected_type_name': 'Dict',
        },
        {
            'original_type_name': 'List[Dict[str, str]]',
            'expected_type_name': 'List',
        },
    )
    def test_get_short_type_name(self, original_type_name, expected_type_name):
        self.assertEqual(
            expected_type_name,
            type_annotations.get_short_type_name(original_type_name))


class TestIsArtifact(parameterized.TestCase):

    @parameterized.parameters([{
        'obj': obj
    } for obj in artifact_types._SCHEMA_TITLE_TO_TYPE.values()])
    def test_true_class(self, obj):
        self.assertTrue(type_annotations.is_artifact_class(obj))

    @parameterized.parameters([{
        'obj': obj(name='name', uri='uri', metadata={})
    } for obj in artifact_types._SCHEMA_TITLE_TO_TYPE.values()])
    def test_true_instance(self, obj):
        self.assertTrue(type_annotations.is_artifact_class(obj))

    @parameterized.parameters([{'obj': 'string'}, {'obj': 1}, {'obj': int}])
    def test_false(self, obj):
        self.assertFalse(type_annotations.is_artifact_class(obj))

    def test_false_no_schema_title(self):

        class NotArtifact:
            schema_version = ''

        self.assertFalse(type_annotations.is_artifact_class(NotArtifact))

    def test_false_no_schema_version(self):

        class NotArtifact:
            schema_title = ''

        self.assertFalse(type_annotations.is_artifact_class(NotArtifact))


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for enabling v2-compatible pipelines in v1."""
import collections
import json
from typing import Optional

from kfp.deprecated import dsl
from kfp.deprecated.compiler import _default_transformers
from kfp.pipeline_spec import pipeline_spec_pb2
from kubernetes import client as k8s_client

_DEFAULT_LAUNCHER_IMAGE = "gcr.io/ml-pipeline/kfp-launcher:1.8.7"


def update_op(op: dsl.ContainerOp,
              pipeline_name: dsl.PipelineParam,
              pipeline_root: dsl.PipelineParam,
              launcher_image: Optional[str] = None) -> None:
    """Updates the passed in Op for running in v2-compatible mode.

    Args:
      op: The Op to update.
      pipeline_spec: The PipelineSpec for the pipeline under which `op`
        runs.
      pipeline_root: The root output directory for pipeline artifacts.
      launcher_image: An optional launcher image. Useful for tests.
    """
    op.is_v2 = True
    # Inject the launcher binary and overwrite the entrypoint.
    image_name = launcher_image or _DEFAULT_LAUNCHER_IMAGE
    launcher_container = dsl.UserContainer(
        name="kfp-launcher",
        image=image_name,
        command=["launcher", "--copy", "/kfp-launcher/launch"],
        mirror_volume_mounts=True)

    op.add_init_container(launcher_container)
    op.add_volume(k8s_client.V1Volume(name='kfp-launcher'))
    op.add_volume_mount(
        k8s_client.V1VolumeMount(
            name='kfp-launcher', mount_path='/kfp-launcher'))

    # op.command + op.args will have the following sections:
    # 1. args passed to kfp-launcher
    # 2. a separator "--"
    # 3. parameters in format "key1=value1", "key2=value2", ...
    # 4. a separator "--" as end of arguments passed to launcher
    # 5. (start of op.args) arguments of the original user program command + args
    #
    # example:
    # - command:
    # - /kfp-launcher/launch
    # - '--mlmd_server_address'
    # - $(METADATA_GRPC_SERVICE_HOST)
    # - '--mlmd_server_port'
    # - $(METADATA_GRPC_SERVICE_PORT)
    # - ... # more launcher params
    # - '--pipeline_task_id'
    # - $(KFP_POD_NAME)
    # - '--pipeline_root'
    # - ''
    # - '--' # start of parameter values
    # - first=first
    # - second=second
    # - '--' # start of user command and args
    # args:
    # - sh
    # - '-ec'
    # - |
    #     program_path=$(mktemp)
    #     printf "%s" "$0" > "$program_path"
    #     python3 -u "$program_path" "$@"
    # - >
    #     import json
    #     import xxx
    #     ...
    op.command = [
        "/kfp-launcher/launch",
        "--mlmd_server_address",
        "$(METADATA_GRPC_SERVICE_HOST)",
        "--mlmd_server_port",
        "$(METADATA_GRPC_SERVICE_PORT)",
        "--runtime_info_json",
        "$(KFP_V2_RUNTIME_INFO)",
        "--container_image",
        "$(KFP_V2_IMAGE)",
        "--task_name",
        op.name,
        "--pipeline_name",
        pipeline_name,
        "--run_id",
        "$(KFP_RUN_ID)",
        "--run_resource",
        "workflows.argoproj.io/$(WORKFLOW_ID)",
        "--namespace",
        "$(KFP_NAMESPACE)",
        "--pod_name",
        "$(KFP_POD_NAME)",
        "--pod_uid",
        "$(KFP_POD_UID)",
        "--pipeline_root",
        pipeline_root,
        "--enable_caching",
        "$(ENABLE_CACHING)",
    ]

    # Mount necessary environment variables.
    op.apply(_default_transformers.add_kfp_pod_env)
    op.container.add_env_variable(
        k8s_client.V1EnvVar(name="KFP_V2_IMAGE", value=op.container.image))

    config_map_ref = k8s_client.V1ConfigMapEnvSource(
        name='metadata-grpc-configmap', optional=True)
    op.container.add_env_from(
        k8s_client.V1EnvFromSource(config_map_ref=config_map_ref))

    op.arguments = list(op.container_spec.command) + list(
        op.container_spec.args)

    runtime_info = {
        "inputParameters": collections.OrderedDict(),
        "inputArtifacts": collections.OrderedDict(),
        "outputParameters": collections.OrderedDict(),
        "outputArtifacts": collections.OrderedDict(),
    }

    op.command += ["--"]
    component_spec = op.component_spec
    for parameter, spec in sorted(
            component_spec.input_definitions.parameters.items()):
        parameter_type = pipeline_spec_pb2.ParameterType.ParameterTypeEnum.Name(
            spec.parameter_type)
        parameter_info = {"type": parameter_type}

        parameter_value = op._parameter_arguments[parameter]
        op.command += [f"{parameter}={parameter_value}"]

        runtime_info["inputParameters"][parameter] = parameter_info
    op.command += ["--"]

    for artifact_name, spec in sorted(
            component_spec.input_definitions.artifacts.items()):
        artifact_info = {
            "metadataPath": op.input_artifact_paths[artifact_name],
            "schemaTitle": spec.artifact_type.schema_title,
            "instanceSchema": spec.artifact_type.instance_schema,
            "schemaVersion": spec.artifact_type.schema_version,
        }
        runtime_info["inputArtifacts"][artifact_name] = artifact_info

    for parameter, spec in sorted(
            component_spec.output_definitions.parameters.items()):
        parameter_info = {
            "type":
                pipeline_spec_pb2.ParameterType.ParameterTypeEnum.Name(
                    spec.parameter_type),
            "path":
                op.file_outputs[parameter],
        }
        runtime_info["outputParameters"][parameter] = parameter_info

    for artifact_name, spec in sorted(
            component_spec.output_definitions.artifacts.items()):
        # TODO: Assert instance_schema.
        artifact_info = {
            # Type used to register output artifacts.
            "schemaTitle": spec.artifact_type.schema_title,
            "instanceSchema": spec.artifact_type.instance_schema,
            "schemaVersion": spec.artifact_type.schema_version,
            # File used to write out the registered artifact ID.
            "metadataPath": op.file_outputs[artifact_name],
        }
        runtime_info["outputArtifacts"][artifact_name] = artifact_info

    op.container.add_env_variable(
        k8s_client.V1EnvVar(
            name="KFP_V2_RUNTIME_INFO", value=json.dumps(runtime_info)))

    op.pod_annotations['pipelines.kubeflow.org/v2_component'] = "true"
    op.pod_labels['pipelines.kubeflow.org/v2_component'] = "true"



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.components.type_annoation_utils."""

import unittest
from absl.testing import parameterized
from typing import Any, Dict, List, Optional

from kfp.deprecated.components import type_annotation_utils


class TypeAnnotationUtilsTest(parameterized.TestCase):

    @parameterized.parameters(
        {
            'original_annotation': str,
            'expected_annotation': str,
        },
        {
            'original_annotation': 'MyCustomType',
            'expected_annotation': 'MyCustomType',
        },
        {
            'original_annotation': List[int],
            'expected_annotation': List[int],
        },
        {
            'original_annotation': Optional[str],
            'expected_annotation': str,
        },
        {
            'original_annotation': Optional[Dict[str, float]],
            'expected_annotation': Dict[str, float],
        },
        {
            'original_annotation': Optional[List[Dict[str, Any]]],
            'expected_annotation': List[Dict[str, Any]],
        },
    )
    def test_maybe_strip_optional_from_annotation(self, original_annotation,
                                                  expected_annotation):
        self.assertEqual(
            expected_annotation,
            type_annotation_utils.maybe_strip_optional_from_annotation(
                original_annotation))

    @parameterized.parameters(
        {
            'original_type_name': 'str',
            'expected_type_name': 'str',
        },
        {
            'original_type_name': 'typing.List[int]',
            'expected_type_name': 'List',
        },
        {
            'original_type_name': 'List[int]',
            'expected_type_name': 'List',
        },
        {
            'original_type_name': 'Dict[str, str]',
            'expected_type_name': 'Dict',
        },
        {
            'original_type_name': 'List[Dict[str, str]]',
            'expected_type_name': 'List',
        },
    )
    def test_get_short_type_name(self, original_type_name, expected_type_name):
        self.assertEqual(
            expected_type_name,
            type_annotation_utils.get_short_type_name(original_type_name))


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes and methods that supports argument for ParallelFor."""

import re
from typing import Any, Dict, List, Optional, Union

from kfp.components import pipeline_channel

ItemList = List[Union[int, float, str, Dict[str, Any]]]


def _get_loop_item_type(type_name: str) -> Optional[str]:
    """Extracts the loop item type.

    This method is used for extract the item type from a collection type.
    For example:

        List[str] -> str
        typing.List[int] -> int
        typing.Sequence[str] -> str
        List -> None
        str -> None

    Args:
        type_name: The collection type name, like `List`, Sequence`, etc.

    Returns:
        The collection item type or None if no match found.
    """
    match = re.match('(typing\.)?(?:\w+)(?:\[(?P<item_type>.+)\])', type_name)
    return match['item_type'].lstrip().rstrip() if match else None


def _get_subvar_type(type_name: str) -> Optional[str]:
    """Extracts the subvar type.

    This method is used for extract the value type from a dictionary type.
    For example:

        Dict[str, int] -> int
        typing.Mapping[str, float] -> float

    Args:
        type_name: The dictionary type.

    Returns:
        The dictionary value type or None if no match found.
    """
    match = re.match(
        '(typing\.)?(?:\w+)(?:\[\s*(?:\w+)\s*,\s*(?P<value_type>.+)\])',
        type_name)
    return match['value_type'].lstrip().rstrip() if match else None


class LoopArgument(pipeline_channel.PipelineParameterChannel):
    """Represents the argument that are looped over in a ParallelFor loop.

    The class shouldn't be instantiated by the end user, rather it is
    created automatically by a ParallelFor ops group.

    To create a LoopArgument instance, use one of its factory methods::

        LoopArgument.from_pipeline_channel(...)
        LoopArgument.from_raw_items(...)


    Attributes:
        items_or_pipeline_channel: The raw items or the PipelineChannel object
            this LoopArgument is associated to.
    """
    LOOP_ITEM_NAME_BASE = 'loop-item'
    LOOP_ITEM_PARAM_NAME_BASE = 'loop-item-param'

    def __init__(
        self,
        items: Union[ItemList, pipeline_channel.PipelineChannel],
        name_code: Optional[str] = None,
        name_override: Optional[str] = None,
        **kwargs,
    ):
        """Initializes a LoopArguments object.

        Args:
            items: List of items to loop over.  If a list of dicts then, all
                dicts must have the same keys and every key must be a legal
                Python variable name.
            name_code: A unique code used to identify these loop arguments.
                Should match the code for the ParallelFor ops_group which created
                these LoopArguments. This prevents parameter name collisions.
            name_override: The override name for PipelineChannel.
            **kwargs: Any other keyword arguments passed down to PipelineChannel.
        """
        if (name_code is None) == (name_override is None):
            raise ValueError(
                'Expect one and only one of `name_code` and `name_override` to '
                'be specified.')

        if name_override is None:
            super().__init__(name=self._make_name(name_code), **kwargs)
        else:
            super().__init__(name=name_override, **kwargs)

        if not isinstance(items,
                          (list, tuple, pipeline_channel.PipelineChannel)):
            raise TypeError(
                f'Expected list, tuple, or PipelineChannel, got {items}.')

        if isinstance(items, tuple):
            items = list(items)

        self.items_or_pipeline_channel = items
        self.is_with_items_loop_argument = not isinstance(
            items, pipeline_channel.PipelineChannel)
        self._referenced_subvars: Dict[str, LoopArgumentVariable] = {}

        if isinstance(items, list) and isinstance(items[0], dict):
            subvar_names = set(items[0].keys())
            # then this block creates loop_arg.variable_a and loop_arg.variable_b
            for subvar_name in subvar_names:
                loop_arg_var = LoopArgumentVariable(
                    loop_argument=self,
                    subvar_name=subvar_name,
                )
                self._referenced_subvars[subvar_name] = loop_arg_var
                setattr(self, subvar_name, loop_arg_var)

    def __getattr__(self, name: str):
        # this is being overridden so that we can access subvariables of the
        # LoopArgument (i.e.: item.a) without knowing the subvariable names ahead
        # of time.

        return self._referenced_subvars.setdefault(
            name, LoopArgumentVariable(
                loop_argument=self,
                subvar_name=name,
            ))

    def _make_name(self, code: str):
        """Makes a name for this loop argument from a unique code."""
        return f'{self.LOOP_ITEM_PARAM_NAME_BASE}-{code}'

    @classmethod
    def from_pipeline_channel(
        cls,
        channel: pipeline_channel.PipelineChannel,
    ) -> 'LoopArgument':
        """Creates a LoopArgument object from a PipelineChannel object."""
        return LoopArgument(
            items=channel,
            name_override=channel.name + '-' + cls.LOOP_ITEM_NAME_BASE,
            task_name=channel.task_name,
            channel_type=_get_loop_item_type(channel.channel_type) or 'String',
        )

    @classmethod
    def from_raw_items(
        cls,
        raw_items: ItemList,
        name_code: str,
    ) -> 'LoopArgument':
        """Creates a LoopArgument object from raw item list."""
        if len(raw_items) == 0:
            raise ValueError('Got an empty item list for loop argument.')

        return LoopArgument(
            items=raw_items,
            name_code=name_code,
            channel_type=type(raw_items[0]).__name__,
        )

    @classmethod
    def name_is_loop_argument(cls, name: str) -> bool:
        """Returns True if the given channel name looks like a loop argument.

        Either it came from a withItems loop item or withParams loop
        item.
        """
        return  ('-' + cls.LOOP_ITEM_NAME_BASE) in name \
          or (cls.LOOP_ITEM_PARAM_NAME_BASE + '-') in name


class LoopArgumentVariable(pipeline_channel.PipelineChannel):
    """Represents a subvariable for a loop argument.

    This is used for cases where we're looping over maps, each of which contains
    several variables. If the user ran:

        with dsl.ParallelFor([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}]) as item:
            ...

    Then there's one LoopArgumentVariable for 'a' and another for 'b'.

    Attributes:
        loop_argument: The original LoopArgument object this subvariable is
          attached to.
        subvar_name: The subvariable name.
    """
    SUBVAR_NAME_DELIMITER = '-subvar-'
    LEGAL_SUBVAR_NAME_REGEX = re.compile(r'^[a-zA-Z_][0-9a-zA-Z_]*$')

    def __init__(
        self,
        loop_argument: LoopArgument,
        subvar_name: str,
    ):
        """Initializes a LoopArgumentVariable instance.

        Args:
            loop_argument: The LoopArgument object this subvariable is based on
                a subvariable to.
            subvar_name: The name of this subvariable, which is the name of the
                dict key that spawned this subvariable.

        Raises:
            ValueError is subvar name is illegal.
        """
        if not self._subvar_name_is_legal(subvar_name):
            raise ValueError(
                f'Tried to create subvariable named {subvar_name}, but that is '
                'not a legal Python variable name.')

        self.subvar_name = subvar_name
        self.loop_argument = loop_argument

        super().__init__(
            name=self._get_name_override(
                loop_arg_name=loop_argument.name,
                subvar_name=subvar_name,
            ),
            task_name=loop_argument.task_name,
            channel_type=_get_subvar_type(loop_argument.channel_type) or
            'String',
        )

    @property
    def items_or_pipeline_channel(
            self) -> Union[ItemList, pipeline_channel.PipelineChannel]:
        """Returns the loop argument items."""
        return self.loop_argument.items_or_pipeline_chanenl

    @property
    def is_with_items_loop_argument(self) -> bool:
        """Whether the loop argument is originated from raw items."""
        return self.loop_argument.is_with_items_loop_argument

    def _subvar_name_is_legal(self, proposed_variable_name: str) -> bool:
        """Returns True if the subvar name is legal."""
        return re.match(self.LEGAL_SUBVAR_NAME_REGEX,
                        proposed_variable_name) is not None

    def _get_name_override(self, loop_arg_name: str, subvar_name: str) -> str:
        """Gets the name.

        Args:
            loop_arg_name: the name of the loop argument parameter that this
              LoopArgumentVariable is attached to.
            subvar_name: The name of this subvariable.

        Returns:
            The name of this loop arg variable.
        """
        return f'{loop_arg_name}{self.SUBVAR_NAME_DELIMITER}{subvar_name}'


class Collected(pipeline_channel.PipelineChannel):
    """For collecting into a list the output from a task in dsl.ParallelFor
    loops.

    Args:
        output: The output of an upstream task within a dsl.ParallelFor loop.

    Example:
      ::

        @dsl.pipeline
        def math_pipeline() -> int:
            with dsl.ParallelFor([1, 2, 3]) as x:
                t = double(num=x)

        return add(nums=dsl.Collected(t.output)).output
    """

    def __init__(
        self,
        output: pipeline_channel.PipelineChannel,
    ) -> None:
        self.output = output
        if isinstance(output, pipeline_channel.PipelineArtifactChannel):
            channel_type = output.channel_type
            self.is_artifact_channel = True
            # we know all dsl.Collected instances are lists, so set to true
            # for type checking, which occurs before dsl.Collected is updated to
            # it's "correct" channel during compilation
            self.is_artifact_list = True
        else:
            channel_type = 'LIST'
            self.is_artifact_channel = False

        super().__init__(
            output.name,
            channel_type=channel_type,
            task_name=output.task_name,
        )



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

from kfp.components import placeholders


class ContainerComponentArtifactChannel:
    """A class for passing in placeholders into container_component decorated
    function."""

    def __init__(self, io_type: str, var_name: str):
        self._io_type = io_type
        self._var_name = var_name

    def __getattr__(
        self, _name: str
    ) -> Union[placeholders.InputUriPlaceholder, placeholders
               .InputPathPlaceholder, placeholders.OutputUriPlaceholder,
               placeholders.OutputPathPlaceholder,
               placeholders.InputMetadataPlaceholder,
               placeholders.OutputMetadataPlaceholder]:
        attr_to_placeholder_dict = {
            'uri': {
                'input': placeholders.InputUriPlaceholder,
                'output': placeholders.OutputUriPlaceholder,
            },
            'path': {
                'input': placeholders.InputPathPlaceholder,
                'output': placeholders.OutputPathPlaceholder,
            },
            'metadata': {
                'input': placeholders.InputMetadataPlaceholder,
                'output': placeholders.OutputMetadataPlaceholder
            },
        }
        if _name not in ['uri', 'path', 'metadata']:
            raise AttributeError(f'Cannot access artifact attribute "{_name}".')
        return attr_to_placeholder_dict[_name][self._io_type](self._var_name)



# Copyright 2018 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys

import click
import typer

from kfp.deprecated._client import Client
from kfp.deprecated.cli.run import run
from kfp.deprecated.cli.recurring_run import recurring_run
from kfp.deprecated.cli.pipeline import pipeline
from kfp.deprecated.cli.diagnose_me_cli import diagnose_me
from kfp.deprecated.cli.experiment import experiment
from kfp.deprecated.cli.output import OutputFormat
from kfp.deprecated.cli import components

_NO_CLIENT_COMMANDS = ['diagnose_me', 'components']


@click.group()
@click.option('--endpoint', help='Endpoint of the KFP API service to connect.')
@click.option('--iap-client-id', help='Client ID for IAP protected endpoint.')
@click.option(
    '-n',
    '--namespace',
    default='kubeflow',
    show_default=True,
    help='Kubernetes namespace to connect to the KFP API.')
@click.option(
    '--other-client-id',
    help='Client ID for IAP protected endpoint to obtain the refresh token.')
@click.option(
    '--other-client-secret',
    help='Client ID for IAP protected endpoint to obtain the refresh token.')
@click.option(
    '--output',
    type=click.Choice(list(map(lambda x: x.name, OutputFormat))),
    default=OutputFormat.table.name,
    show_default=True,
    help='The formatting style for command output.')
@click.pass_context
def cli(ctx: click.Context, endpoint: str, iap_client_id: str, namespace: str,
        other_client_id: str, other_client_secret: str, output: OutputFormat):
    """kfp is the command line interface to KFP service.

    Feature stage:
    [Alpha](https://github.com/kubeflow/pipelines/blob/07328e5094ac2981d3059314cc848fbb71437a76/docs/release/feature-stages.md#alpha)
    """
    if ctx.invoked_subcommand in _NO_CLIENT_COMMANDS:
        # Do not create a client for these subcommands
        return
    ctx.obj['client'] = Client(endpoint, iap_client_id, namespace,
                               other_client_id, other_client_secret)
    ctx.obj['namespace'] = namespace
    ctx.obj['output'] = output


def main():
    logging.basicConfig(format='%(message)s', level=logging.INFO)
    cli.add_command(run)
    cli.add_command(recurring_run)
    cli.add_command(pipeline)
    cli.add_command(diagnose_me, 'diagnose_me')
    cli.add_command(experiment)
    cli.add_command(typer.main.get_command(components.app))
    try:
        cli(obj={}, auto_envvar_prefix='KFP')
    except Exception as e:
        click.echo(str(e), err=True)
        sys.exit(1)



# Copyright 2018-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import time
from typing import List

import click
from kfp import client
from kfp.cli import output
from kfp.cli.utils import deprecated_alias_group
from kfp.cli.utils import parsing


@click.group(
    cls=deprecated_alias_group.deprecated_alias_group_factory(
        {'submit': 'create'}))
def run():
    """Manage run resources."""


@run.command()
@click.option(
    '-e',
    '--experiment-id',
    help=parsing.get_param_descr(client.Client.list_runs, 'experiment_id'))
@click.option(
    '--page-token',
    default='',
    help=parsing.get_param_descr(client.Client.list_runs, 'page_token'))
@click.option(
    '-m',
    '--max-size',
    default=100,
    help=parsing.get_param_descr(client.Client.list_runs, 'page_size'))
@click.option(
    '--sort-by',
    default='created_at desc',
    help=parsing.get_param_descr(client.Client.list_runs, 'sort_by'))
@click.option(
    '--filter', help=parsing.get_param_descr(client.Client.list_runs, 'filter'))
@click.pass_context
def list(ctx: click.Context, experiment_id: str, page_token: str, max_size: int,
         sort_by: str, filter: str):
    """List pipeline runs."""
    client_obj: client.Client = ctx.obj['client']
    output_format = ctx.obj['output']
    response = client_obj.list_runs(
        experiment_id=experiment_id,
        page_token=page_token,
        page_size=max_size,
        sort_by=sort_by,
        filter=filter)
    output.print_output(
        response.runs or [],
        output.ModelType.RUN,
        output_format,
    )


@run.command()
@click.option(
    '-e',
    '--experiment-name',
    required=True,
    help='Experiment name of the run.')
@click.option(
    '-r',
    '--run-name',
    help=parsing.get_param_descr(client.Client.run_pipeline, 'job_name'))
@click.option(
    '-f',
    '--package-file',
    type=click.Path(exists=True, dir_okay=False),
    help=parsing.get_param_descr(client.Client.run_pipeline,
                                 'pipeline_package_path'))
@click.option(
    '-p',
    '--pipeline-id',
    help=parsing.get_param_descr(client.Client.run_pipeline, 'pipeline_id'))
@click.option('-n', '--pipeline-name', help='Name of the pipeline template.')
@click.option(
    '-w',
    '--watch',
    is_flag=True,
    default=False,
    help='Watch the run status until it finishes.')
@click.option(
    '-v',
    '--version',
    help=parsing.get_param_descr(client.Client.run_pipeline, 'version_id'))
@click.option(
    '-t',
    '--timeout',
    default=0,
    help='Wait for a run to complete until timeout in seconds.',
    type=int)
@click.argument('args', nargs=-1)
@click.pass_context
def create(ctx: click.Context, experiment_name: str, run_name: str,
           package_file: str, pipeline_id: str, pipeline_name: str, watch: bool,
           timeout: int, version: str, args: List[str]):
    """Submit a pipeline run."""
    client_obj: client.Client = ctx.obj['client']
    namespace = ctx.obj['namespace']
    output_format = ctx.obj['output']
    if not run_name:
        run_name = experiment_name

    if not pipeline_id and pipeline_name:
        pipeline_id = client_obj.get_pipeline_id(name=pipeline_name)

    if not package_file and not pipeline_id and not version:
        click.echo(
            'You must provide one of [package_file, pipeline_id, version].',
            err=True)
        sys.exit(1)

    arg_dict = dict(arg.split('=', maxsplit=1) for arg in args)

    experiment = client_obj.create_experiment(experiment_name)
    run = client_obj.run_pipeline(
        experiment_id=experiment.experiment_id,
        job_name=run_name,
        pipeline_package_path=package_file,
        params=arg_dict,
        pipeline_id=pipeline_id,
        version_id=version)
    if timeout > 0:
        run_detail = client_obj.wait_for_run_completion(run.run_id, timeout)
        output.print_output(
            run_detail,
            output.ModelType.RUN,
            output_format,
        )
    else:
        display_run(client_obj, run.run_id, watch, output_format)


@run.command()
@click.option(
    '-w',
    '--watch',
    is_flag=True,
    default=False,
    help='Watch the run status until it finishes.')
@click.option(
    '-d',
    '--detail',
    is_flag=True,
    default=False,
    help='Get detailed information of the run in json format.')
@click.argument('run-id')
@click.pass_context
def get(ctx: click.Context, watch: bool, detail: bool, run_id: str):
    """Get information about a pipeline run."""
    client_obj: client.Client = ctx.obj['client']
    namespace = ctx.obj['namespace']
    output_format = ctx.obj['output']
    if detail:
        output_format = 'json'
        click.echo(
            'The --detail/-d flag is deprecated. Please use --output=json instead.',
            err=True)
    display_run(client_obj, run_id, watch, output_format)


@run.command()
@click.argument('run-id')
@click.pass_context
def archive(ctx: click.Context, run_id: str):
    """Archive a pipeline run."""
    client_obj: client.Client = ctx.obj['client']
    output_format = ctx.obj['output']

    client_obj.archive_run(run_id=run_id)
    run = client_obj.get_run(run_id=run_id)
    output.print_output(
        run,
        output.ModelType.RUN,
        output_format,
    )


@run.command()
@click.argument('run-id')
@click.pass_context
def unarchive(ctx: click.Context, run_id: str):
    """Unarchive a pipeline run."""
    client_obj: client.Client = ctx.obj['client']
    output_format = ctx.obj['output']
    client_obj.unarchive_run(run_id=run_id)
    run = client_obj.get_run(run_id=run_id)
    output.print_output(
        run,
        output.ModelType.RUN,
        output_format,
    )


@run.command()
@click.argument('run-id')
@click.pass_context
def delete(ctx: click.Context, run_id: str):
    """Delete a pipeline run."""

    confirmation = f'Are you sure you want to delete run {run_id}?'
    if not click.confirm(confirmation):
        return

    client_obj: client.Client = ctx.obj['client']
    output_format = ctx.obj['output']

    client_obj.delete_run(run_id=run_id)
    output.print_deleted_text('run', run_id, output_format)


def display_run(client: client.Client, run_id: str, watch: bool,
                output_format: str):
    run = client.get_run(run_id)

    output.print_output(
        run,
        output.ModelType.RUN,
        output_format,
    )
    if not watch:
        return
    while True:
        time.sleep(1)
        run_detail = client.get_run(run_id)
        run = run_detail
        if run_detail.state in [
                'SUCCEEDED', 'SKIPPED', 'FAILED', 'CANCELED', 'PAUSED'
        ]:
            click.echo(f'Run is finished with state {run_detail.state}.')
            return



# Copyright 2018 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import google.auth
import google.auth.app_engine
import google.auth.compute_engine.credentials
import google.auth.iam
from google.auth.transport.requests import Request
import google.oauth2.credentials
import google.oauth2.service_account
import requests_toolbelt.adapters.appengine
from webbrowser import open_new_tab
import requests
import json

IAM_SCOPE = 'https://www.googleapis.com/auth/iam'
OAUTH_TOKEN_URI = 'https://www.googleapis.com/oauth2/v4/token'
LOCAL_KFP_CREDENTIAL = os.path.expanduser('~/.config/kfp/credentials.json')


def get_gcp_access_token():
    """Get and return GCP access token for the current Application Default
    Credentials.

    If not set, returns None. For more information, see
    https://cloud.google.com/sdk/gcloud/reference/auth/application-default/print-access-token
    """
    token = None
    try:
        creds, project = google.auth.default(
            scopes=["https://www.googleapis.com/auth/cloud-platform"])
        if not creds.valid:
            auth_req = Request()
            creds.refresh(auth_req)
        if creds.valid:
            token = creds.token
    except Exception as e:
        logging.warning('Failed to get GCP access token: %s', e)
    return token


def get_auth_token(client_id, other_client_id, other_client_secret):
    """Gets auth token from default service account or user account."""
    if os.path.exists(LOCAL_KFP_CREDENTIAL):
        # fetch IAP auth token using the locally stored credentials.
        with open(LOCAL_KFP_CREDENTIAL, 'r') as f:
            credentials = json.load(f)
        if client_id in credentials:
            return id_token_from_refresh_token(
                credentials[client_id]['other_client_id'],
                credentials[client_id]['other_client_secret'],
                credentials[client_id]['refresh_token'], client_id)
    if other_client_id is None or other_client_secret is None:
        # fetch IAP auth token: service accounts
        token = get_auth_token_from_sa(client_id)
    else:
        # fetch IAP auth token: user account
        # Obtain the ID token for provided Client ID with user accounts.
        #  Flow: get authorization code -> exchange for refresh token -> obtain and return ID token
        refresh_token = get_refresh_token_from_client_id(
            other_client_id, other_client_secret)
        credentials = {}
        if os.path.exists(LOCAL_KFP_CREDENTIAL):
            with open(LOCAL_KFP_CREDENTIAL, 'r') as f:
                credentials = json.load(f)
        credentials[client_id] = {}
        credentials[client_id]['other_client_id'] = other_client_id
        credentials[client_id]['other_client_secret'] = other_client_secret
        credentials[client_id]['refresh_token'] = refresh_token
        #TODO: handle the case when the refresh_token expires.
        #   which only happens if the refresh_token is not used once for six months.
        if not os.path.exists(os.path.dirname(LOCAL_KFP_CREDENTIAL)):
            os.makedirs(os.path.dirname(LOCAL_KFP_CREDENTIAL))
        with open(LOCAL_KFP_CREDENTIAL, 'w') as f:
            json.dump(credentials, f)
        token = id_token_from_refresh_token(other_client_id,
                                            other_client_secret, refresh_token,
                                            client_id)
    return token


def get_auth_token_from_sa(client_id):
    """Gets auth token from default service account.

    If no service account credential is found, returns None.
    """
    service_account_credentials = get_service_account_credentials(client_id)
    if service_account_credentials:
        return get_google_open_id_connect_token(service_account_credentials)
    return None


def get_service_account_credentials(client_id):
    # Figure out what environment we're running in and get some preliminary
    # information about the service account.
    bootstrap_credentials, _ = google.auth.default(scopes=[IAM_SCOPE])
    if isinstance(bootstrap_credentials, google.oauth2.credentials.Credentials):
        logging.info('Found OAuth2 credentials and skip SA auth.')
        return None
    elif isinstance(bootstrap_credentials, google.auth.app_engine.Credentials):
        requests_toolbelt.adapters.appengine.monkeypatch()

    # For service account's using the Compute Engine metadata service,
    # service_account_email isn't available until refresh is called.
    bootstrap_credentials.refresh(Request())
    signer_email = bootstrap_credentials.service_account_email
    if isinstance(bootstrap_credentials,
                  google.auth.compute_engine.credentials.Credentials):
        # Since the Compute Engine metadata service doesn't expose the service
        # account key, we use the IAM signBlob API to sign instead.
        # In order for this to work:
        #
        # 1. Your VM needs the https://www.googleapis.com/auth/iam scope.
        #    You can specify this specific scope when creating a VM
        #    through the API or gcloud. When using Cloud Console,
        #    you'll need to specify the "full access to all Cloud APIs"
        #    scope. A VM's scopes can only be specified at creation time.
        #
        # 2. The VM's default service account needs the "Service Account Actor"
        #    role. This can be found under the "Project" category in Cloud
        #    Console, or roles/iam.serviceAccountActor in gcloud.
        signer = google.auth.iam.Signer(Request(), bootstrap_credentials,
                                        signer_email)
    else:
        # A Signer object can sign a JWT using the service account's key.
        signer = bootstrap_credentials.signer

    # Construct OAuth 2.0 service account credentials using the signer
    # and email acquired from the bootstrap credentials.
    return google.oauth2.service_account.Credentials(
        signer,
        signer_email,
        token_uri=OAUTH_TOKEN_URI,
        additional_claims={'target_audience': client_id})


def get_google_open_id_connect_token(service_account_credentials):
    """Get an OpenID Connect token issued by Google for the service account.

    This function:
      1. Generates a JWT signed with the service account's private key
         containing a special "target_audience" claim.
      2. Sends it to the OAUTH_TOKEN_URI endpoint. Because the JWT in #1
         has a target_audience claim, that endpoint will respond with
         an OpenID Connect token for the service account -- in other words,
         a JWT signed by *Google*. The aud claim in this JWT will be
         set to the value from the target_audience claim in #1.
    For more information, see
    https://developers.google.com/identity/protocols/OAuth2ServiceAccount .
    The HTTP/REST example on that page describes the JWT structure and
    demonstrates how to call the token endpoint. (The example on that page
    shows how to get an OAuth2 access token; this code is using a
    modified version of it to get an OpenID Connect token.)
    """

    service_account_jwt = (
        service_account_credentials._make_authorization_grant_assertion())
    request = google.auth.transport.requests.Request()
    body = {
        'assertion': service_account_jwt,
        'grant_type': google.oauth2._client._JWT_GRANT_TYPE,
    }
    token_response = google.oauth2._client._token_endpoint_request(
        request, OAUTH_TOKEN_URI, body)
    return token_response['id_token']


def get_refresh_token_from_client_id(client_id, client_secret):
    """Obtain the ID token for provided Client ID with user accounts.

    Flow: get authorization code -> exchange for refresh token -> obtain and return ID token
    """
    auth_code = get_auth_code(client_id)
    return get_refresh_token_from_code(auth_code, client_id, client_secret)


def get_auth_code(client_id):
    auth_url = "https://accounts.google.com/o/oauth2/v2/auth?client_id=%s&response_type=code&scope=openid%%20email&access_type=offline&redirect_uri=urn:ietf:wg:oauth:2.0:oob" % client_id
    print(auth_url)
    open_new_tab(auth_url)
    return input(
        "If there's no browser window prompt, please direct to the URL above, then copy and paste the authorization code here: "
    )


def get_refresh_token_from_code(auth_code, client_id, client_secret):
    payload = {
        "code": auth_code,
        "client_id": client_id,
        "client_secret": client_secret,
        "redirect_uri": "urn:ietf:wg:oauth:2.0:oob",
        "grant_type": "authorization_code"
    }
    res = requests.post(OAUTH_TOKEN_URI, data=payload)
    res.raise_for_status()
    return str(json.loads(res.text)[u"refresh_token"])


def id_token_from_refresh_token(client_id, client_secret, refresh_token,
                                audience):
    payload = {
        "client_id": client_id,
        "client_secret": client_secret,
        "refresh_token": refresh_token,
        "grant_type": "refresh_token",
        "audience": audience
    }
    res = requests.post(OAUTH_TOKEN_URI, data=payload)
    res.raise_for_status()
    return str(json.loads(res.text)[u"id_token"])



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Deprecated. See kfp.components.types.type_utils instead.

This module will be removed in KFP v2.0.
"""
import warnings
from kfp.components.types import type_utils
from kfp.pipeline_spec import pipeline_spec_pb2
import inspect
from typing import Union, Type
import re
from typing import List, Optional
from kfp.deprecated.components import _structures
from kfp.components.types import artifact_types

warnings.warn(
    'Module kfp.dsl.type_utils is deprecated and will be removed'
    ' in KFP v2.0. Please use from kfp.components.types.type_utils instead.',
    category=FutureWarning)

is_parameter_type = type_utils.is_parameter_type
get_parameter_type = type_utils.get_parameter_type

# copying a lot of code from v2 to here to avoid certain dependencies of deprecated on v2, since making any changes to this code in v2 would result in breaks in deprecated/
_GOOGLE_TYPES_PATTERN = r'^google.[A-Za-z]+$'
_GOOGLE_TYPES_VERSION = '0.0.1'
_ARTIFACT_CLASSES_MAPPING = {
    'model': artifact_types.Model,
    'dataset': artifact_types.Dataset,
    'metrics': artifact_types.Metrics,
    'classificationmetrics': artifact_types.ClassificationMetrics,
    'slicedclassificationmetrics': artifact_types.SlicedClassificationMetrics,
    'html': artifact_types.HTML,
    'markdown': artifact_types.Markdown,
}


def get_artifact_type_schema(
    artifact_class_or_type_name: Optional[Union[str,
                                                Type[artifact_types.Artifact]]]
) -> pipeline_spec_pb2.ArtifactTypeSchema:
    """Gets the IR I/O artifact type msg for the given ComponentSpec I/O
    type."""
    artifact_class = artifact_types.Artifact
    if isinstance(artifact_class_or_type_name, str):
        if re.match(_GOOGLE_TYPES_PATTERN, artifact_class_or_type_name):
            return pipeline_spec_pb2.ArtifactTypeSchema(
                schema_title=artifact_class_or_type_name,
                schema_version=_GOOGLE_TYPES_VERSION,
            )
        artifact_class = _ARTIFACT_CLASSES_MAPPING.get(
            artifact_class_or_type_name.lower(), artifact_types.Artifact)
    elif inspect.isclass(artifact_class_or_type_name) and issubclass(
            artifact_class_or_type_name, artifact_types.Artifact):
        artifact_class = artifact_class_or_type_name

    return pipeline_spec_pb2.ArtifactTypeSchema(
        schema_title=artifact_class.schema_title,
        schema_version=artifact_class.schema_version)


def get_input_artifact_type_schema(
    input_name: str,
    inputs: List[_structures.InputSpec],
) -> Optional[str]:
    """Find the input artifact type by input name.

    Args:
      input_name: The name of the component input.
      inputs: The list of InputSpec

    Returns:
      The artifact type schema of the input.

    Raises:
      AssertionError if input not found, or input found but not an artifact type.
    """

    for component_input in inputs:
        if component_input.name == input_name:
            assert not is_parameter_type(
                component_input.type), 'Input is not an artifact type.'
            return get_artifact_type_schema_old(component_input.type)
    assert False, 'Input not found.'



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

from kfp.components import component_factory
from kfp.components import container_component


def container_component(
        func: Callable) -> container_component.ContainerComponent:
    """Decorator for container-based components in KFP v2.

    Args:
        func: The python function to create a component from. The function
            should have type annotations for all its arguments, indicating how
            it is intended to be used (e.g. as an input/output Artifact object,
            a plain parameter, or a path to a file).

    Example:
      ::

        from kfp.dsl import container_component, ContainerSpec, InputPath, OutputPath, Output

        @container_component
        def my_component(
            dataset_path: InputPath(Dataset),
            model: Output[Model],
            num_epochs: int,
            output_parameter: OutputPath(str),
        ):
            return ContainerSpec(
                image='gcr.io/my-image',
                command=['sh', 'my_component.sh'],
                args=[
                '--dataset_path', dataset_path,
                '--model_path', model.path,
                '--output_parameter_path', output_parameter,
            ]
        )
    """
    return component_factory.create_container_component_from_func(func)



# Copyright 2018 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import copy
import json
import pathlib
from typing import Any, Mapping, Optional

from kfp.deprecated._config import COMPILING_FOR_V2
from kfp.deprecated.components import _structures
from kfp.deprecated.components import _components
from kfp.deprecated.components import _naming
from kfp.deprecated import dsl
from kfp.deprecated.dsl import _container_op
from kfp.deprecated.dsl import _for_loop
from kfp.deprecated.dsl import _pipeline_param
from kfp.deprecated.dsl import component_spec as dsl_component_spec
from kfp.deprecated.dsl import dsl_utils
from kfp.deprecated.dsl import types
from kfp.pipeline_spec import pipeline_spec_pb2
from kfp.components.types import type_utils

# Placeholder to represent the output directory hosting all the generated URIs.
# Its actual value will be specified during pipeline compilation.
# The format of OUTPUT_DIR_PLACEHOLDER is serialized dsl.PipelineParam, to
# ensure being extracted as a pipeline parameter during compilation.
# Note that we cannot direclty import dsl module here due to circular
# dependencies.
OUTPUT_DIR_PLACEHOLDER = '{{pipelineparam:op=;name=pipeline-root}}'
# Placeholder to represent to UID of the current pipeline at runtime.
# Will be replaced by engine-specific placeholder during compilation.
RUN_ID_PLACEHOLDER = '{{kfp.run_uid}}'
# Format of the Argo parameter used to pass the producer's Pod ID to
# the consumer.
PRODUCER_POD_NAME_PARAMETER = '{}-producer-pod-id-'
# Format of the input output port name placeholder.
INPUT_OUTPUT_NAME_PATTERN = '{{{{kfp.input-output-name.{}}}}}'
# Fixed name for per-task output metadata json file.
OUTPUT_METADATA_JSON = '/tmp/outputs/executor_output.json'
# Executor input placeholder.
_EXECUTOR_INPUT_PLACEHOLDER = '{{$}}'


# TODO(chensun): block URI placeholder usage in v1.
def _generate_output_uri_placeholder(port_name: str) -> str:
    """Generates the URI placeholder for an output."""
    return "{{{{$.outputs.artifacts['{}'].uri}}}}".format(port_name)


def _generate_input_uri_placeholder(port_name: str) -> str:
    """Generates the URI placeholder for an input."""
    return "{{{{$.inputs.artifacts['{}'].uri}}}}".format(port_name)


def _generate_output_metadata_path() -> str:
    """Generates the URI to write the output metadata JSON file."""

    return OUTPUT_METADATA_JSON


def _generate_input_metadata_path(port_name: str) -> str:
    """Generates the placeholder for input artifact metadata file."""

    # Return a placeholder for path to input artifact metadata, which will be
    # rewritten during pipeline compilation.
    return str(
        pathlib.PurePosixPath(
            OUTPUT_DIR_PLACEHOLDER, RUN_ID_PLACEHOLDER,
            '{{{{inputs.parameters.{input}}}}}'.format(
                input=PRODUCER_POD_NAME_PARAMETER.format(port_name)),
            OUTPUT_METADATA_JSON))


def _generate_input_output_name(port_name: str) -> str:
    """Generates the placeholder for input artifact's output name."""

    # Return a placeholder for the output port name of the input artifact, which
    # will be rewritten during pipeline compilation.
    return INPUT_OUTPUT_NAME_PATTERN.format(port_name)


def _generate_executor_input() -> str:
    """Generates the placeholder for serialized executor input."""
    return _EXECUTOR_INPUT_PLACEHOLDER


class ExtraPlaceholderResolver:

    def __init__(self):
        self.input_paths = {}
        self.input_metadata_paths = {}
        self.output_paths = {}

    def resolve_placeholder(
        self,
        arg,
        component_spec: _structures.ComponentSpec,
        arguments: dict,
    ) -> str:
        inputs_dict = {
            input_spec.name: input_spec
            for input_spec in component_spec.inputs or []
        }

        if isinstance(arg, _structures.InputUriPlaceholder):
            input_name = arg.input_name
            if input_name in arguments:
                input_uri = _generate_input_uri_placeholder(input_name)
                self.input_paths[
                    input_name] = _components._generate_input_file_name(
                        input_name)
                return input_uri
            else:
                input_spec = inputs_dict[input_name]
                if input_spec.optional:
                    return None
                else:
                    raise ValueError(
                        'No value provided for input {}'.format(input_name))

        elif isinstance(arg, _structures.OutputUriPlaceholder):
            output_name = arg.output_name
            output_uri = _generate_output_uri_placeholder(output_name)
            self.output_paths[
                output_name] = _components._generate_output_file_name(
                    output_name)
            return output_uri

        elif isinstance(arg, _structures.InputMetadataPlaceholder):
            input_name = arg.input_name
            if input_name in arguments:
                input_metadata_path = _generate_input_metadata_path(input_name)
                self.input_metadata_paths[input_name] = input_metadata_path
                return input_metadata_path
            else:
                input_spec = inputs_dict[input_name]
                if input_spec.optional:
                    return None
                else:
                    raise ValueError(
                        'No value provided for input {}'.format(input_name))

        elif isinstance(arg, _structures.InputOutputPortNamePlaceholder):
            input_name = arg.input_name
            if input_name in arguments:
                return _generate_input_output_name(input_name)
            else:
                input_spec = inputs_dict[input_name]
                if input_spec.optional:
                    return None
                else:
                    raise ValueError(
                        'No value provided for input {}'.format(input_name))

        elif isinstance(arg, _structures.OutputMetadataPlaceholder):
            # TODO: Consider making the output metadata per-artifact.
            return _generate_output_metadata_path()
        elif isinstance(arg, _structures.ExecutorInputPlaceholder):
            return _generate_executor_input()

        return None


def _create_container_op_from_component_and_arguments(
    component_spec: _structures.ComponentSpec,
    arguments: Mapping[str, Any],
    component_ref: Optional[_structures.ComponentReference] = None,
) -> _container_op.ContainerOp:
    """Instantiates ContainerOp object.

    Args:
      component_spec: The component spec object.
      arguments: The dictionary of component arguments.
      component_ref: (only for v1) The component references.

    Returns:
      A ContainerOp instance.
    """
    # Add component inputs with default value to the arguments dict if they are not
    # in the arguments dict already.
    arguments = arguments.copy()
    for input_spec in component_spec.inputs or []:
        if input_spec.name not in arguments and input_spec.default is not None:
            default_value = input_spec.default
            if input_spec.type == 'Integer':
                default_value = int(default_value)
            elif input_spec.type == 'Float':
                default_value = float(default_value)
            elif (type_utils.is_parameter_type(input_spec.type) and
                  COMPILING_FOR_V2):
                parameter_type = type_utils.get_parameter_type(input_spec.type)
                default_value = type_utils.deserialize_parameter_value(
                    value=default_value, parameter_type=parameter_type)

            arguments[input_spec.name] = default_value

    # Check types of the reference arguments and serialize PipelineParams
    original_arguments = arguments
    arguments = arguments.copy()
    for input_name, argument_value in arguments.items():
        if isinstance(argument_value, _pipeline_param.PipelineParam):
            input_type = component_spec._inputs_dict[input_name].type
            argument_type = argument_value.param_type
            types.verify_type_compatibility(
                argument_type, input_type,
                'Incompatible argument passed to the input "{}" of component "{}": '
                .format(input_name, component_spec.name))

            arguments[input_name] = str(argument_value)
        if isinstance(argument_value, _container_op.ContainerOp):
            raise TypeError(
                'ContainerOp object was passed to component as an input argument. '
                'Pass a single output instead.')
    placeholder_resolver = ExtraPlaceholderResolver()
    resolved_cmd = _components._resolve_command_line_and_paths(
        component_spec=component_spec,
        arguments=arguments,
        placeholder_resolver=placeholder_resolver.resolve_placeholder,
    )

    container_spec = component_spec.implementation.container

    old_warn_value = _container_op.ContainerOp._DISABLE_REUSABLE_COMPONENT_WARNING
    _container_op.ContainerOp._DISABLE_REUSABLE_COMPONENT_WARNING = True

    output_paths = collections.OrderedDict(resolved_cmd.output_paths or {})
    output_paths.update(placeholder_resolver.output_paths)
    input_paths = collections.OrderedDict(resolved_cmd.input_paths or {})
    input_paths.update(placeholder_resolver.input_paths)

    artifact_argument_paths = [
        dsl.InputArgumentPath(
            argument=arguments[input_name],
            input=input_name,
            path=path,
        ) for input_name, path in input_paths.items()
    ]

    task = _container_op.ContainerOp(
        name=component_spec.name or _components._default_component_name,
        image=container_spec.image,
        command=resolved_cmd.command,
        arguments=resolved_cmd.args,
        file_outputs=output_paths,
        artifact_argument_paths=artifact_argument_paths,
    )
    _container_op.ContainerOp._DISABLE_REUSABLE_COMPONENT_WARNING = old_warn_value

    component_meta = copy.copy(component_spec)
    task._set_metadata(component_meta, original_arguments)
    if component_ref:
        component_ref_without_spec = copy.copy(component_ref)
        component_ref_without_spec.spec = None
        task._component_ref = component_ref_without_spec

    task._parameter_arguments = resolved_cmd.inputs_consumed_by_value
    name_to_spec_type = {}
    if component_meta.inputs:
        name_to_spec_type = {
            input.name: {
                'type': input.type,
                'default': input.default,
            } for input in component_meta.inputs
        }

    if COMPILING_FOR_V2:
        for name, spec_type in name_to_spec_type.items():
            if (name in original_arguments and
                    type_utils.is_parameter_type(spec_type['type'])):
                if isinstance(original_arguments[name], (list, dict)):
                    task._parameter_arguments[name] = json.dumps(
                        original_arguments[name])
                else:
                    task._parameter_arguments[name] = str(
                        original_arguments[name])

    for name in list(task.artifact_arguments.keys()):
        if name in task._parameter_arguments:
            del task.artifact_arguments[name]

    for name in list(task.input_artifact_paths.keys()):
        if name in task._parameter_arguments:
            del task.input_artifact_paths[name]

    # Previously, ContainerOp had strict requirements for the output names, so we
    # had to convert all the names before passing them to the ContainerOp
    # constructor.
    # Outputs with non-pythonic names could not be accessed using their original
    # names. Now ContainerOp supports any output names, so we're now using the
    # original output names. However to support legacy pipelines, we're also
    # adding output references with pythonic names.
    # TODO: Add warning when people use the legacy output names.
    output_names = [
        output_spec.name for output_spec in component_spec.outputs or []
    ]  # Stabilizing the ordering
    output_name_to_python = _naming.generate_unique_name_conversion_table(
        output_names, _naming._sanitize_python_function_name)
    for output_name in output_names:
        pythonic_output_name = output_name_to_python[output_name]
        # Note: Some component outputs are currently missing from task.outputs
        # (e.g. MLPipeline UI Metadata)
        if pythonic_output_name not in task.outputs and output_name in task.outputs:
            task.outputs[pythonic_output_name] = task.outputs[output_name]

    if container_spec.env:
        from kubernetes import client as k8s_client
        for name, value in container_spec.env.items():
            task.container.add_env_variable(
                k8s_client.V1EnvVar(name=name, value=value))

    if component_spec.metadata:
        annotations = component_spec.metadata.annotations or {}
        for key, value in annotations.items():
            task.add_pod_annotation(key, value)
        for key, value in (component_spec.metadata.labels or {}).items():
            task.add_pod_label(key, value)
        # Disabling the caching for the volatile components by default
        if annotations.get('volatile_component', 'false') == 'true':
            task.execution_options.caching_strategy.max_cache_staleness = 'P0D'

    _attach_v2_specs(task, component_spec, original_arguments)

    return task


def _attach_v2_specs(
    task: _container_op.ContainerOp,
    component_spec: _structures.ComponentSpec,
    arguments: Mapping[str, Any],
) -> None:
    """Attaches v2 specs to a ContainerOp object.

    Attach v2_specs to the ContainerOp object regardless whether the pipeline is
    being compiled to v1 (Argo yaml) or v2 (IR json).
    However, there're different behaviors for the two cases. Namely, resolved
    commands and arguments, error handling, etc.
    Regarding the difference in error handling, v2 has a stricter requirement on
    input type annotation. For instance, an input without any type annotation is
    viewed as an artifact, and if it's paired with InputValuePlaceholder, an
    error will be thrown at compile time. However, we cannot raise such an error
    in v1, as it wouldn't break existing pipelines.

    Args:
      task: The ContainerOp object to attach IR specs.
      component_spec: The component spec object.
      arguments: The dictionary of component arguments.
    """

    def _resolve_commands_and_args_v2(
        component_spec: _structures.ComponentSpec,
        arguments: Mapping[str, Any],
    ) -> _components._ResolvedCommandLineAndPaths:
        """Resolves the command line argument placeholders for v2 (IR).

        Args:
          component_spec: The component spec object.
          arguments: The dictionary of component arguments.

        Returns:
          A named tuple: _components._ResolvedCommandLineAndPaths.
        """
        inputs_dict = {
            input_spec.name: input_spec
            for input_spec in component_spec.inputs or []
        }
        outputs_dict = {
            output_spec.name: output_spec
            for output_spec in component_spec.outputs or []
        }

        def _input_artifact_uri_placeholder(input_key: str) -> str:
            if COMPILING_FOR_V2 and type_utils.is_parameter_type(
                    inputs_dict[input_key].type):
                raise TypeError(
                    'Input "{}" with type "{}" cannot be paired with '
                    'InputUriPlaceholder.'.format(input_key,
                                                  inputs_dict[input_key].type))
            else:
                return _generate_input_uri_placeholder(input_key)

        def _input_artifact_path_placeholder(input_key: str) -> str:
            if COMPILING_FOR_V2 and type_utils.is_parameter_type(
                    inputs_dict[input_key].type):
                raise TypeError(
                    'Input "{}" with type "{}" cannot be paired with '
                    'InputPathPlaceholder.'.format(input_key,
                                                   inputs_dict[input_key].type))
            else:
                return "{{{{$.inputs.artifacts['{}'].path}}}}".format(input_key)

        def _input_parameter_placeholder(input_key: str) -> str:
            if COMPILING_FOR_V2 and not type_utils.is_parameter_type(
                    inputs_dict[input_key].type):
                raise TypeError(
                    'Input "{}" with type "{}" cannot be paired with '
                    'InputValuePlaceholder.'.format(
                        input_key, inputs_dict[input_key].type))
            else:
                return "{{{{$.inputs.parameters['{}']}}}}".format(input_key)

        def _output_artifact_uri_placeholder(output_key: str) -> str:
            if COMPILING_FOR_V2 and type_utils.is_parameter_type(
                    outputs_dict[output_key].type):
                raise TypeError(
                    'Output "{}" with type "{}" cannot be paired with '
                    'OutputUriPlaceholder.'.format(
                        output_key, outputs_dict[output_key].type))
            else:
                return _generate_output_uri_placeholder(output_key)

        def _output_artifact_path_placeholder(output_key: str) -> str:
            return "{{{{$.outputs.artifacts['{}'].path}}}}".format(output_key)

        def _output_parameter_path_placeholder(output_key: str) -> str:
            return "{{{{$.outputs.parameters['{}'].output_file}}}}".format(
                output_key)

        def _resolve_output_path_placeholder(output_key: str) -> str:
            if type_utils.is_parameter_type(outputs_dict[output_key].type):
                return _output_parameter_path_placeholder(output_key)
            else:
                return _output_artifact_path_placeholder(output_key)

        placeholder_resolver = ExtraPlaceholderResolver()

        def _resolve_ir_placeholders_v2(
            arg,
            component_spec: _structures.ComponentSpec,
            arguments: dict,
        ) -> str:
            inputs_dict = {
                input_spec.name: input_spec
                for input_spec in component_spec.inputs or []
            }
            if isinstance(arg, _structures.InputValuePlaceholder):
                input_name = arg.input_name
                input_value = arguments.get(input_name, None)
                if input_value is not None:
                    return _input_parameter_placeholder(input_name)
                else:
                    input_spec = inputs_dict[input_name]
                    if input_spec.optional:
                        return None
                    else:
                        raise ValueError(
                            'No value provided for input {}'.format(input_name))

            elif isinstance(arg, _structures.InputUriPlaceholder):
                input_name = arg.input_name
                if input_name in arguments:
                    input_uri = _input_artifact_uri_placeholder(input_name)
                    return input_uri
                else:
                    input_spec = inputs_dict[input_name]
                    if input_spec.optional:
                        return None
                    else:
                        raise ValueError(
                            'No value provided for input {}'.format(input_name))

            elif isinstance(arg, _structures.OutputUriPlaceholder):
                output_name = arg.output_name
                output_uri = _output_artifact_uri_placeholder(output_name)
                return output_uri

            return placeholder_resolver.resolve_placeholder(
                arg=arg,
                component_spec=component_spec,
                arguments=arguments,
            )

        resolved_cmd = _components._resolve_command_line_and_paths(
            component_spec=component_spec,
            arguments=arguments,
            input_path_generator=_input_artifact_path_placeholder,
            output_path_generator=_resolve_output_path_placeholder,
            placeholder_resolver=_resolve_ir_placeholders_v2,
        )
        return resolved_cmd

    pipeline_task_spec = pipeline_spec_pb2.PipelineTaskSpec()

    # Check types of the reference arguments and serialize PipelineParams
    arguments = arguments.copy()

    # Preserve input params for ContainerOp.inputs
    input_params_set = set([
        param for param in arguments.values()
        if isinstance(param, _pipeline_param.PipelineParam)
    ])

    for input_name, argument_value in arguments.items():
        input_type = component_spec._inputs_dict[input_name].type
        argument_type = None

        if isinstance(argument_value, _pipeline_param.PipelineParam):
            argument_type = argument_value.param_type

            types.verify_type_compatibility(
                argument_type, input_type,
                'Incompatible argument passed to the input "{}" of component "{}": '
                .format(input_name, component_spec.name))

            # Loop arguments defaults to 'String' type if type is unknown.
            # This has to be done after the type compatiblity check.
            if argument_type is None and isinstance(
                    argument_value,
                (_for_loop.LoopArguments, _for_loop.LoopArgumentVariable)):
                argument_type = 'String'

            arguments[input_name] = str(argument_value)

            if type_utils.is_parameter_type(input_type):
                if argument_value.op_name:
                    pipeline_task_spec.inputs.parameters[
                        input_name].task_output_parameter.producer_task = (
                            dsl_utils.sanitize_task_name(
                                argument_value.op_name))
                    pipeline_task_spec.inputs.parameters[
                        input_name].task_output_parameter.output_parameter_key = (
                            argument_value.name)
                else:
                    pipeline_task_spec.inputs.parameters[
                        input_name].component_input_parameter = argument_value.name
            else:
                if argument_value.op_name:
                    pipeline_task_spec.inputs.artifacts[
                        input_name].task_output_artifact.producer_task = (
                            dsl_utils.sanitize_task_name(
                                argument_value.op_name))
                    pipeline_task_spec.inputs.artifacts[
                        input_name].task_output_artifact.output_artifact_key = (
                            argument_value.name)
        elif isinstance(argument_value, str):
            argument_type = 'String'
            pipeline_params = _pipeline_param.extract_pipelineparams_from_any(
                argument_value)
            if pipeline_params and COMPILING_FOR_V2:
                # argument_value contains PipelineParam placeholders which needs to be
                # replaced. And the input needs to be added to the task spec.
                for param in pipeline_params:
                    # Form the name for the compiler injected input, and make sure it
                    # doesn't collide with any existing input names.
                    additional_input_name = (
                        dsl_component_spec
                        .additional_input_name_for_pipelineparam(param))
                    for existing_input_name, _ in arguments.items():
                        if existing_input_name == additional_input_name:
                            raise ValueError(
                                'Name collision between existing input name '
                                '{} and compiler injected input name {}'.format(
                                    existing_input_name, additional_input_name))

                    # Add the additional param to the input params set. Otherwise, it will
                    # not be included when the params set is not empty.
                    input_params_set.add(param)
                    additional_input_placeholder = (
                        "{{{{$.inputs.parameters['{}']}}}}".format(
                            additional_input_name))
                    argument_value = argument_value.replace(
                        param.pattern, additional_input_placeholder)

                    # The output references are subject to change -- the producer task may
                    # not be whitin the same DAG.
                    if param.op_name:
                        pipeline_task_spec.inputs.parameters[
                            additional_input_name].task_output_parameter.producer_task = (
                                dsl_utils.sanitize_task_name(param.op_name))
                        pipeline_task_spec.inputs.parameters[
                            additional_input_name].task_output_parameter.output_parameter_key = param.name
                    else:
                        pipeline_task_spec.inputs.parameters[
                            additional_input_name].component_input_parameter = param.full_name

            input_type = component_spec._inputs_dict[input_name].type
            if type_utils.is_parameter_type(input_type):
                pipeline_task_spec.inputs.parameters[
                    input_name].runtime_value.constant.string_value = argument_value
        elif isinstance(argument_value, int):
            argument_type = 'Integer'
            pipeline_task_spec.inputs.parameters[
                input_name].runtime_value.constant.number_value = argument_value
        elif isinstance(argument_value, float):
            argument_type = 'Float'
            pipeline_task_spec.inputs.parameters[
                input_name].runtime_value.constant.number_value = argument_value
        elif isinstance(argument_value, bool):
            argument_type = 'Bool'
            pipeline_task_spec.inputs.parameters[
                input_name].runtime_value.constant.bool_value = argument_value
        elif isinstance(argument_value, list):
            argument_type = 'List'

            # Convert any PipelineParams to strings.
            argument_value = map(
                lambda x: str(x)
                if isinstance(x, dsl.PipelineParam) else x, argument_value)

            pipeline_task_spec.inputs.parameters[
                input_name].runtime_value.constant.list_value.extend(
                    argument_value)
        elif isinstance(argument_value, dict):
            argument_type = 'Dict'
            pipeline_task_spec.inputs.parameters[
                input_name].runtime_value.constant.struct_value.update(
                    argument_value)
        elif isinstance(argument_value, _container_op.ContainerOp):
            raise TypeError(
                f'ContainerOp object {input_name} was passed to component as an '
                'input argument. Pass a single output instead.')
        else:
            if COMPILING_FOR_V2:
                raise NotImplementedError(
                    'Input argument supports only the following types: '
                    'PipelineParam, str, int, float, bool, dict, and list. Got: '
                    f'"{argument_value}".')

        argument_is_parameter_type = type_utils.is_parameter_type(argument_type)
        input_is_parameter_type = type_utils.is_parameter_type(input_type)
        if COMPILING_FOR_V2 and (argument_is_parameter_type !=
                                 input_is_parameter_type):
            if isinstance(argument_value, dsl.PipelineParam):
                param_or_value_msg = 'PipelineParam "{}"'.format(
                    argument_value.full_name)
            else:
                param_or_value_msg = 'value "{}"'.format(argument_value)

            raise TypeError(
                'Passing '
                '{param_or_value} with type "{arg_type}" (as "{arg_category}") to '
                'component input '
                '"{input_name}" with type "{input_type}" (as "{input_category}") is '
                'incompatible. Please fix the type of the component input.'
                .format(
                    param_or_value=param_or_value_msg,
                    arg_type=argument_type,
                    arg_category='Parameter'
                    if argument_is_parameter_type else 'Artifact',
                    input_name=input_name,
                    input_type=input_type,
                    input_category='Parameter'
                    if input_is_parameter_type else 'Artifact',
                ))

    if not component_spec.name:
        component_spec.name = _components._default_component_name

    resolved_cmd = _resolve_commands_and_args_v2(
        component_spec=component_spec, arguments=arguments)

    task.container_spec = (
        pipeline_spec_pb2.PipelineDeploymentConfig.PipelineContainerSpec(
            image=component_spec.implementation.container.image,
            command=resolved_cmd.command,
            args=resolved_cmd.args,
            env=[
                pipeline_spec_pb2.PipelineDeploymentConfig.PipelineContainerSpec
                .EnvVar(name=name, value=value)
                for name, value in task.container.env_dict.items()
            ],
        ))

    # TODO(chensun): dedupe IR component_spec and contaienr_spec
    pipeline_task_spec.component_ref.name = (
        dsl_utils.sanitize_component_name(task.name))
    executor_label = dsl_utils.sanitize_executor_label(task.name)

    task.component_spec = dsl_component_spec.build_component_spec_from_structure(
        component_spec, executor_label, arguments.keys())

    task.task_spec = pipeline_task_spec

    # Override command and arguments if compiling to v2.
    if COMPILING_FOR_V2:
        task.command = resolved_cmd.command
        task.arguments = resolved_cmd.args

        # limit this to v2 compiling only to avoid possible behavior change in v1.
        task.inputs = list(input_params_set)



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
import unittest

from kfp.client import set_volume_credentials
from kfp.client import token_credentials_base
from kubernetes.client import configuration


class TestServiceAccountTokenVolumeCredentials(unittest.TestCase):

    def test_get_token(self):
        with tempfile.TemporaryDirectory() as tempdir:
            token_path = os.path.join(tempdir, 'token.txt')
            with open(token_path, 'w') as f:
                f.write('my_token')
            service_account_tvc = set_volume_credentials.ServiceAccountTokenVolumeCredentials(
                token_path)

            self.assertEqual(service_account_tvc._get_token(), 'my_token')

    def test_refresh_api_key_hook(self):
        with tempfile.TemporaryDirectory() as tempdir:
            token_path = os.path.join(tempdir, 'token.txt')
            with open(token_path, 'w') as f:
                f.write('my_token')
            service_account_tvc = set_volume_credentials.ServiceAccountTokenVolumeCredentials(
                token_path)
            config = configuration.Configuration()

            with self.assertRaisesRegex(KeyError, r'authorization'):
                config.api_key['authorization']

            service_account_tvc.refresh_api_key_hook(config)
            self.assertEqual(config.api_key['authorization'], 'my_token')



# Copyright 2018-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import warnings

from kfp.components import v1_structures
import yaml


def _load_component_spec_from_component_text(
        text) -> v1_structures.ComponentSpec:
    component_dict = yaml.safe_load(text)
    component_spec = v1_structures.ComponentSpec.from_dict(component_dict)

    if isinstance(component_spec.implementation,
                  v1_structures.ContainerImplementation) and (
                      component_spec.implementation.container.command is None):
        warnings.warn(
            'Container component must specify command to be compatible with KFP '
            'v2 compatible mode and emissary executor, which will be the default'
            ' executor for KFP v2.'
            'https://www.kubeflow.org/docs/components/pipelines/installation/choose-executor/',
            category=FutureWarning,
        )

    # Calculating hash digest for the component
    data = text if isinstance(text, bytes) else text.encode('utf-8')
    data = data.replace(b'\r\n', b'\n')  # Normalizing line endings
    digest = hashlib.sha256(data).hexdigest()
    component_spec._digest = digest

    return component_spec
